# -*- coding: utf-8 -*-
"""
本模块功能：单项证券的VaR(在险价值)和ES(预期损失)计算函数包
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2018年10月10日
最新修订日期：2019年10月10日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
作者邮件：wdehong2000@163.com
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""
#==============================================================================
#统一屏蔽一般性警告
import warnings; warnings.filterwarnings("ignore")     
from siat.common import *
from siat.translate import *
#==============================================================================
import matplotlib.pyplot as plt

#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
    mpfrc={'font.family': 'SimHei'}

if czxt in ['darwin']: #MacOSX
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family': 'Heiti TC'}

if czxt in ['linux']: #website Jupyter
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family':'Heiti TC'}

# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================

def get_stock_quotes(ticker,start_date,end_date):
    """
    输入参数：股票代码，开始日期，结束日期
    输出参数：
    股票价格序列(日期，开盘价，最高价，最低价，收盘价，调整收盘价，交易量)	
    start_date, end_date是datetime类型
    """

    import pandas_datareader as pdr
    try:
        stock_quotes=pdr.DataReader(ticker,'yahoo',start_date,end_date)
    except: return None
    if len(stock_quotes) == 0: return None
    
    return stock_quotes    
    # stock_quotes是dataframe类型

#==============================================================================
def get_end_price(stock_quotes):
    """
    输入参数：股票价格序列
    输出参数：最新股价(金额)
    """
    end_price=stock_quotes['Close'][-1]
    return end_price

#==============================================================================
def get_ret_series(stock_quotes):
    """
    输入参数：股票价格序列
    输出参数：股票日收益率序列(注意不是DataFrame)
    """
    stock_quotes['ret']=stock_quotes['Close'].pct_change()
    stock_quotes=stock_quotes.dropna()
    ret_series=stock_quotes['ret']  
    return ret_series
    # ret_series是序列类型

#==============================================================================
def VaR_normal_standard(position,ret_series,future_days=1,alpha=0.99):
    
    """
    标准正太法VaR基本算法
    输入参数：当前持有头寸金额，收益率序列(非百分比)，未来持有时间(天)，置信度
    输出参数：VaR(金额，单位与当前头寸的金额单位相同)，负数
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import numpy as np
    from scipy import stats
    
    z=stats.norm.ppf(1-alpha)
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    ratio=miu_days+z*sigma_days
    VaR_days=position*ratio

    return VaR_days

#==============================================================================
def stock_VaR_normal_standard(ticker,shares,today, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：持有股票的VaR，标准正态法
    输入参数：股票代码，持有股数，当前日期，未来持有时间(天)，置信度，使用历史数据的年数
    输出：VaR(负数金额，单位与股价的金额单位相同)，VaR比率
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if p is None: 
        print("#Error(stock_VaR_normal_standard): no obs retrieved.")        
        return None,None
    
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end
    VaR=VaR_normal_standard(position,r,future_days,alpha)
    if VaR is None: return None,None
    #最大为全损
    if abs(VaR) > position: VaR=-position
    
    VaR_ratio=abs(VaR/position)
    
    if printout == True:
        print("=== 计算在险价值：标准正态模型 ===")
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有日期    :",today)
        print("预计持有天数:",future_days)
        print("置信度      : ",alpha*100,'%',sep='')
        print("在险价值VaR :",format(round(VaR,2),','))
        print("VaR比率     : ",round(VaR_ratio*100,2),'%',sep='')
        
        import datetime as dt; today=dt.date.today()   
        footnote="*数据来源：雅虎财经，"+str(today)
        print(footnote)
    
    return VaR,VaR_ratio

if __name__ == '__main__':
    var1,ratio1=stock_VaR_normal_standard('BABA',10000,'2019-08-08',1,0.99)

#==============================================================================
def series_VaR_normal_standard(ticker,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：一步计算在多个日期持有一定天数股票资产的VaR
    输入参数：股票代码，持有股数，日期列表(日期型列表)，未来持有时间(天数)，置信度，使用历史数据的年数
    输出：多个日期的VaR(日期，股票代码，VaR金额，VaR比率(即单位头寸的VaR))
    # datelist是datetime类型日期的列表
    """

    import pandas as pd
    result=pd.DataFrame(columns=['date','ticker','VaR','ratio'])
    for d in datelist:
        VaR,ratio=stock_VaR_normal_standard(ticker,shares,d, \
                future_days,alpha,pastyears,printout=False)
        if (VaR is None) or (ratio is None): continue
        
        s = pd.Series({'date':d,'ticker':ticker,'VaR':VaR,'ratio':ratio})
        result=result.append(s,ignore_index=True)
    result2=result.set_index(['date'])    
    # result2是dateframe类型
    
    if printout == False: return result2
    
    #打印
    result3=result2.copy()
    result3['VaR金额']=round(result3['VaR'],2)
    result3['VaR比例%']=round(result3['ratio']*100,2)
    result3.drop(columns=['ticker','VaR','ratio'],inplace=True)
    
    text1="= "+codetranslate(ticker)+": VaR比例，持有"+str(future_days)+"天 ="
    print(text1)    
    print(result3)
    import datetime as dt; today=dt.date.today()    
    print("*数据来源：雅虎财经，"+str(today))    
    
    #绘图    
    #VaR金额绘图    
    plt.plot(result3['VaR金额'],c='r',lw=2)  
    title1=codetranslate(ticker)+": VaR金额的变化，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('VaR金额')  
    plt.xticks(rotation=45)
    
    import datetime as dt; today=dt.date.today()   
    footnote="数据来源：雅虎财经，"+str(today)    
    plt.xlabel(footnote)
    plt.show()    
    
    plt.plot(result3['VaR比例%'],c='r',lw=2)  
    title2=codetranslate(ticker)+": VaR比例的变化， 持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('VaR比例%')  
    plt.xticks(rotation=45)
    plt.xlabel(footnote)
    plt.show()
    
    return result2

if __name__ == '__main__':
    datelist=['2018-01-01','2018-04-01','2018-07-01','2018-10-01', \
              '2019-01-01','2019-04-01','2019-07-01']
    result=series_VaR_normal_standard('BABA',10000,datelist,1,0.99)


#==============================================================================
def compare_VaR_normal_standard(tickerlist,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1):
    """
    功能：比较多个日期持有一定天数股票资产的VaR高低
    输入参数：股票列表，持有股数，日期列表(日期型列表)，未来持有时间(天数)，置信度，使用历史数据的年数
    输出：无
    显示：折线图，各个资产的VaR金额和比率对比
    """
    
    print("... The comparison may take time, please wait ...")
    markerlist=['.','o','s','*','+','x','1','2']
    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]         
        r=series_VaR_normal_standard(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['VaR amount']=round(rr['VaR'],2)
        rr.drop(columns=['ticker','VaR','ratio'],inplace=True)        
        plt.plot(rr['VaR amount'],label=codetranslate(t),lw=2,marker=thismarker)

    title1="比较VaR金额，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('VaR金额')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    
    import datetime as dt; today=dt.date.today()   
    footnote="数据来源：雅虎财经，"+str(today)    
    plt.xlabel(footnote)    
    plt.show() 
    
    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]        
        r=series_VaR_normal_standard(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['VaR ratio %']=round(rr['ratio']*100,2)
        rr.drop(columns=['ticker','VaR','ratio'],inplace=True)        
        plt.plot(rr['VaR ratio %'],label=codetranslate(t),lw=2,marker=thismarker)

    title2="比较VaR比例，持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('VaR比例%')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    plt.xlabel(footnote)
    plt.show() 

    return 

if __name__ == '__main__':
    tickerlist=['BABA','PDD','JD']
    datelist=['2019-01-01','2019-02-01','2019-03-01','2019-04-01', \
              '2019-05-01','2019-06-01','2019-07-01']
    compare_VaR_normal_standard(tickerlist,10000,datelist,1,0.99)

#==============================================================================
def ES_normal_standard(position,ret_series,future_days=1,alpha=0.99):
    """
    功能：计算ES，标准正态法
    输入参数：持有头寸金额，日收益率序列，未来持有日期，置信度
    输出：预期损失(金额)
    """
    import numpy as np
    from scipy import stats
    
    z=stats.norm.ppf(1-alpha)
    miu_daily=np.mean(ret_series)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(ret_series)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    zES=-stats.norm.pdf(z)/(1-alpha)
    ratio=miu_days+zES*sigma_days
    ES_days=position*ratio
    
    return -ES_days

#==============================================================================
def stock_ES_normal_standard(ticker,shares,today, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：计算持有股票资产的ES，标准正态法
    输入参数：股票代码，持有股数，当前日期，未来持有日期(天数)，置信度，使用历史数据的年数
    #输出参数：预期损失(金额)负数，预期损失与头寸的比率
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)
    if (p is None) or (len(p)==0): 
        print("#Error(stock_ES_normal_standard): no obs retrieved.")                
        return None,None
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end
    ES=ES_normal_standard(position,r,future_days,alpha)
    #最大为全损
    if abs(ES) > position: ES=-position
    
    ratio=abs(ES/position)
    
    if printout == True:
        print("=== 计算预期不足ES：标准正态模型 ===")
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有日期    :",today)
        print("预计持有天数:",future_days)
        print("置信度      : ",alpha*100,'%',sep='')
        print("ES金额     :",format(round(ES,2),','))
        print("ES比例     : ",round(ratio*100,2),'%',sep='')    
        
        import datetime as dt; today=dt.date.today()
        print("*数据来源：雅虎财经，"+str(today))
    
    return ES,ratio


#==============================================================================
def series_ES_normal_standard(ticker,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：一步计算在多个日期持有一定天数股票资产的ES，标准正态法
    输入参数：股票代码，持有股数，日期列表(日期型列表)，未来持有时间(天数)，置信度，使用历史数据的年数
    输出：多个日期的ES(日期，股票代码，ES金额，ES比率(即单位头寸的ES))
    # datelist是datetime类型日期的列表
    """

    import pandas as pd
    result=pd.DataFrame(columns=['date','ticker','ES','ratio'])
    for d in datelist:
        ES,ratio=stock_ES_normal_standard(ticker,shares,d, \
                future_days,alpha,pastyears,printout=False)
        if (ES is None) or (ratio is None): continue
        
        s = pd.Series({'date':d,'ticker':ticker,'ES':ES,'ratio':ratio})
        result=result.append(s,ignore_index=True)
    result2=result.set_index(['date'])    
    # result2是dateframe类型
    
    if printout == False: return result2
    
    #打印
    result3=result2.copy()
    result3['ES金额']=round(result3['ES'],2)
    result3['ES比例%']=round(result3['ratio']*100,2)
    result3.drop(columns=['ticker','ES','ratio'],inplace=True)
    
    text1="=== "+codetranslate(ticker)+": ES比例，持有"+str(future_days)+"天 ==="
    print(text1)    
    print(result3)
    import datetime as dt; today=dt.date.today()    
    print("*数据来源：雅虎财经，"+str(today))
    
    #绘图    
    #VaR金额绘图    
    plt.plot(result3['ES金额'],c='r',lw=2)  
    title1=codetranslate(ticker)+": ES金额的变化，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('ES金额')  
    plt.xticks(rotation=45)
    
    plt.show()    
    
    plt.plot(result3['EES比例%'],c='r',lw=2)  
    title2=codetranslate(ticker)+": ES比例的变化，持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('ES比例%')  
    plt.xticks(rotation=45)
    
    plt.show()
    
    return result2

if __name__ == '__main__':
    datelist=['2018-01-01','2018-04-01','2018-07-01','2018-10-01', \
              '2019-01-01','2019-04-01','2019-07-01']
    result=series_ES_normal_standard('BABA',10000,datelist,1,0.99)


#==============================================================================
def compare_ES_normal_standard(tickerlist,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1):
    """
    功能：比较多个日期持有一定天数股票资产的ES高低，标准正态法
    输入参数：股票列表，持有股数，日期列表(日期型列表)，未来持有时间(天数)，置信度，使用历史数据的年数
    输出：无
    显示：折线图，各个资产的ES金额和比率对比
    """
    
    print("... The comparison may take time, please wait ...")
    markerlist=['.','o','s','*','+','x','1','2']
    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]        
        r=series_ES_normal_standard(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['ES amount']=round(rr['ES'],2)
        rr.drop(columns=['ticker','ES','ratio'],inplace=True)        
        plt.plot(rr['ES amount'],label=codetranslate(t),lw=2,marker=thismarker)

    title1="比较ES金额，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('ES金额')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    
    import datetime as dt; today=dt.date.today() 
    footnote="数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)    
    plt.show() 

    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]        
        r=series_ES_normal_standard(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['ES ratio %']=round(rr['ratio']*100,2)
        rr.drop(columns=['ticker','ES','ratio'],inplace=True)        
        plt.plot(rr['ES ratio %'],label=codetranslate(t),lw=2,marker=thismarker)

    title2="比较ES比例，持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('ES比例%')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    plt.xlabel(footnote)
    plt.show() 

    return 

if __name__ == '__main__':
    tickerlist=['BABA','PDD','JD']
    datelist=['2019-01-01','2019-02-01','2019-03-01','2019-04-01', \
              '2019-05-01','2019-06-01','2019-07-01']
    compare_ES_normal_standard(tickerlist,10000,datelist,1,0.99)


#==============================================================================
def normfunc(x,mu,sigma):
    #计算正态分布的概率密度，带有均值和标准差
    import numpy as np
    pdf = np.exp(-((x - mu)**2)/(2*sigma**2)) / (sigma * np.sqrt(2*np.pi))
    return pdf

#==============================================================================
def plot_rets_histogram(ticker,start,end,num_bins=20):
    """
    功能：绘制收益率分布的直方图，并于相应的正态分布图对照
    输入：股票代码，开始/结束时间
    显示：收益率分布的直方图(实线)，相应的正态分布图(虚线)
    x轴为收益率(非百分比)，y轴为频度(Frequency)
    """
    #抓取股价并计算收益率
    quotes=get_stock_quotes(ticker,start,end)
    if (quotes is None) or (len(quotes)==0):
        print("#Error(plot_rets_histogram): Fetching data failed")
        print("Information:",ticker,start,end)
        return    
    rets=get_ret_series(quotes)

    #计算收益率的均值和标准差
    mu=rets.mean()
    sigma=rets.std()
    
    #绘制股票收益率直方图
    plt.figure(figsize=(8,4))
    n,bins,patches=plt.hist(rets,num_bins,facecolor='blue',alpha=0.5,label=codetranslate(ticker))
    
    #生成与直方图柱子对应的正态分布概率密度
    y=normfunc(bins,mu,sigma)
    #绘制正态分布曲线
    plt.plot(bins,y,'r--',label='正态分布',lw=2)
    plt.ylabel('Frequency')
    plt.xlabel('Stock return')
    titletxt="正态性检验："+codetranslate(ticker)+"股票收益率, "+start+"至"+end
    plt.title(titletxt)
    plt.legend(loc='best')
    
    import datetime as dt; today=dt.date.today() 
    footnote="数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)    
    plt.show()    
    
    return

if __name__ == '__main__':
    plot_rets_histogram('JD','2019-1-1','2019-6-30')

#==============================================================================
def plot_rets_curve(ticker,start,end):
    """
    功能：绘制收益率分布的曲线，并于相应的正态分布图对照
    输入：股票代码，开始/结束时间
    显示：收益率分布的直方图(实线)，相应的正态分布图(虚线)
    x轴为收益率(非百分比)，y轴为频度(Frequency)
    """
    #抓取股价并计算收益率
    quotes=get_stock_quotes(ticker,start,end)
    if (quotes is None) or (len(quotes)==0):
        print("#Error(plot_rets_curve): Fetching data failed")
        print("Information:",ticker,start,end)
        return
    rets=get_ret_series(quotes)

    #计算收益率的均值和标准差
    mu=rets.mean()
    sigma=rets.std()

    #生成符合正态分布的随机数，符合股票收益率的均值和标准差
    import numpy as np
    x=mu+sigma*np.random.randn(1000000)    
        
    plt.figure(figsize=(8,4))
    import seaborn as sns
    #绘制曲线：股票收益率
    sns.kdeplot(data=rets,shade=True,color='blue',legend=True,label=codetranslate(ticker),lw=2) 
    #绘制曲线：对应的正态分布
    sns.kdeplot(data=x,shade=True,color='r',legend=True,label='正态分布',ls='--')
    #设置标题、图例、坐标轴标签
    #plt.ylabel('观察个数')
    plt.xlabel('收益率')
    plt.legend(loc='best')
    titletxt="正态性检验: "+codetranslate(ticker)+"股票收益率, "+start+"至"+end
    plt.title(titletxt)
    
    import datetime as dt; today=dt.date.today() 
    footnote="数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)
    plt.show()

    return
    

if __name__ == '__main__':
    plot_rets_curve('JD','2019-1-1','2019-6-30')


#===========================================================================
def stock_ret_Normality_SW(ticker,start_date,end_date,siglevel=0.05):
    """
    功能：测试一个日收益率序列是否符合正态分布，原假设：符合正态分布
    输入参数：股票代码，开始日期，结束日期
    输出：日收益率序列正态性检验的p-value
    start_date,end_date均为datetime类型
    【Shapiro-Wilk正态性检验】原假设：服从正态分布
    """
    from scipy import stats
    
    quotes=get_stock_quotes(ticker,start_date,end_date)
    if quotes is None: return None
    if len(quotes) == 0: return None
    
    ret=get_ret_series(quotes)
    (W,p_value)=stats.shapiro(ret)
    
    print("= Shapiro-Wilk正态性检验: 股票收益率 =")
    print("股票  ：",codetranslate(ticker)) 
    print("期间  ：",start_date,"至",end_date)
    print("原假设： 符合正态分布")
    print("W值   ：",round(W,4))
    print("p值   ：",round(p_value,4))
    if p_value >= siglevel:
        print("结果  ： 接受原假设, 符合正态分布")
    else:
        print("结果  ： 拒绝原假设, 不符合正态分布")
    import datetime as dt; today=dt.date.today()
    print("*数据来源：雅虎财经，"+str(today))
    
    return p_value
    
#==============================================================================
def VaR_normal_modified(position,ret_series,future_days=1,alpha=0.99):
    """
    功能：VaR基本算法，修正正态法
    #输入参数：持有头寸金额，日收益率序列，未来持有日期，置信度
    #输出参数：VaR(金额)
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    from scipy import stats
    import numpy as np

    z=np.abs(stats.norm.ppf(1-alpha))
    S=stats.skew(r)
    K=stats.kurtosis(r)
    
    t1=1/6*(np.power(z,2)-1)*S
    t2=1/24*(np.power(z,3)-3*z)*K
    t3=1/36*(2*np.power(z,3)-5*z)*np.power(S,2)
    t=z+t1+t2-t3
    
    miu_daily=np.mean(r)
    miu_days=np.power(miu_daily+1,future_days)-1
    sigma_daily=np.std(r)    
    sigma_days=np.sqrt(future_days)*sigma_daily
    
    ratio=miu_days+t*sigma_days
    VaR_days=position*ratio
    
    return -abs(VaR_days)

#==============================================================================
def stock_VaR_normal_modified(ticker,shares,today, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：计算持有一定量股票若干天的VaR，修正正态法
    输入参数：股票代码，持有股数，当前日期，未来持有天数，置信度，使用历史数据的年数
    注：当前日期可以为过去的任意一天，历史年数为使用几年的历史数据来分析
    输出：VaR(金额，负数)，VaR比率
    注释：VaR比率，即每单位持有金额的在险价值
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if (p is None) or (len(p)==0): 
        print("#Error(stock_VaR_normal_modified): no obs retrieved.")                        
        return None,None
    
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end
    VaR=VaR_normal_modified(position,r,future_days,alpha)
    #最大为全损
    if abs(VaR) > position: VaR=-position
    
    VaR_ratio=abs(VaR/position)
    
    if printout == True:
        print("=== 计算在险价值VaR： 修正正态模型 ===")
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有日期    :",today)
        print("预计持有天数:",future_days)
        print("置信度      : ",alpha*100,'%',sep='')
        print("VaR金额     :",format(round(VaR,2),','))
        print("VaR比例     : ",round(VaR_ratio*100,2),'%',sep='')    
        
        import datetime as dt; today=dt.date.today()
        print("*数据来源：雅虎财经，"+str(today))
    
    return VaR,VaR_ratio    
    
#==============================================================================
def series_VaR_normal_modified(ticker,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：一步计算多个日期的VaR，修正正态分布法
    输入参数：股票代码，持有股数，日期列表，未来持有天数，置信度，使用历史数据的年数
    注释：当前日期可以为过去的任意一天，历史年数为使用几年的历史数据来分析
    输出：VaR信息表(日期，股票代码，VaR金额，VaR比率)
    """
    # datelist是datetime类型日期的列表
    import pandas as pd
    result=pd.DataFrame(columns=['date','ticker','VaR','ratio'])
    for d in datelist:
        VaR,ratio=stock_VaR_normal_modified(ticker,shares,d, \
                future_days,alpha,pastyears,printout=False)
        s = pd.Series({'date':d,'ticker':ticker,'VaR':VaR,'ratio':ratio})
        result=result.append(s,ignore_index=True)
    result2=result.set_index(['date'])    
    # result2是dateframe类型
    
    if printout == False: return result2
    
    #打印
    result3=result2.copy()
    result3['VaR金额']=round(result3['VaR'],2)
    result3['VaR比例%']=round(result3['ratio']*100,2)
    result3.drop(columns=['ticker','VaR','ratio'],inplace=True)
    
    text1="= "+codetranslate(ticker)+": VaR比例，持有"+str(future_days)+"天 ="
    print(text1)    
    print(result3)
    import datetime as dt; today=dt.date.today()
    print("*数据来源：雅虎财经，"+str(today))
    
    #绘图    
    #VaR金额绘图    
    plt.plot(result3['VaR金额'],c='r',lw=2)  
    title1=codetranslate(ticker)+": VaR金额的变化，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('VaR金额')  
    plt.xticks(rotation=45)
    
    footnote="数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)
    plt.show()    
    
    plt.plot(result3['VaR比例%'],c='r',lw=2)  
    title2=codetranslate(ticker)+": VaR比例的变化，持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('VaR比例%')  
    plt.xticks(rotation=45)
    plt.xlabel(footnote)
    plt.show()

    return result2


#==============================================================================
def compare_VaR_normal_modified(tickerlist,shares,datelist, \
                future_days=1,alpha=0.99,pastyears=1):
    """
    功能：比较多个日期持有一定天数股票资产的VaR高低，修正正态法
    输入参数：股票列表，持有股数，日期列表(日期型列表)，未来持有时间(天数)，置信度，使用历史数据的年数
    输出：无
    显示：折线图，各个资产的VaR金额和比率对比
    """
    
    print("... The comparison may take time, please wait ...")
    markerlist=['.','o','s','*','+','x','1','2']
    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]            
        r=series_VaR_normal_modified(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['VaR amount']=round(rr['VaR'],2)
        rr.drop(columns=['ticker','VaR','ratio'],inplace=True)        
        plt.plot(rr['VaR amount'],label=codetranslate(t),lw=2,marker=thismarker)

    title1="比较VaR金额，持有"+str(future_days)+"天"
    plt.title(title1)
    plt.ylabel('VaR金额')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    
    import datetime as dt; today=dt.date.today() 
    footnote="数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)
    plt.show() 

    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]        
        r=series_VaR_normal_modified(t,shares,datelist, \
                future_days,alpha,pastyears,printout=False)
        rr=r.copy()
        rr['VaR ratio %']=round(rr['ratio']*100,2)
        rr.drop(columns=['ticker','VaR','ratio'],inplace=True)        
        plt.plot(rr['VaR ratio %'],label=codetranslate(t),lw=2,marker=thismarker)

    title2="比较VaR比例，持有"+str(future_days)+"天"
    plt.title(title2)
    plt.ylabel('VaR比例%')  
    plt.xticks(rotation=45)
    plt.legend(loc='best')
    plt.xlabel(footnote)
    plt.show() 

    return 

if __name__ == '__main__':
    tickerlist=['BABA','PDD','JD']
    datelist=['2019-01-01','2019-02-01','2019-03-01','2019-04-01', \
              '2019-05-01','2019-06-01','2019-07-01']
    compare_VaR_normal_standard(tickerlist,10000,datelist,1,0.99)


#==============================================================================
def VaR_historical_1d(position,ret_series,alpha=0.99):
    """
    功能：计算VaR基本算法，历史模拟法，持有1天
    输入参数：持有头寸金额，历史日收益率序列，置信度
    输出：持有一天的VaR(金额)
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import numpy as np
    n=len(r)
    t=int(n*(1-alpha))
    SR=np.sort(r)
    if t>=1:
        A=SR[t-1]   #SR的第一个元素的序号是0
    else:
        A=SR[0]
    VaR_1d=position*A
    
    return -abs(VaR_1d)


#==============================================================================
def stock_VaR_historical_1d(ticker,shares,today,alpha=0.99, \
                            pastyears=1,printout=True):
    """
    功能：计算持有股票的VaR，历史模拟法，持有1天
    输入参数：股票代码，持有股数，当前日期，置信度，使用历史数据的年数
    输出：持有一天的VaR(金额和比率)
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if (p is None) or (len(p)==0): 
        print("#Error(stock_VaR_historical_1d): no obs retrieved.")        
        return None,None
    
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end
    VaR_1d=VaR_historical_1d(position,r,alpha)
    #最高为全损
    if abs(VaR_1d)>position: VaR_1d=-position
    VaR_ratio=abs(VaR_1d/position)

    if printout == True:
        print("=== 计算在险价值VaR：历史模拟方法 ===")
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有日期    :",today)
        future_days=1
        print("预计持有日期:",future_days)
        print("置信度     : ",alpha*100,'%',sep='')
        print("VaR金额    :",format(round(VaR_1d,2),','))
        print("VaR比例    : ",round(VaR_ratio*100,2),'%',sep='')    
        
        import datetime as dt; today=dt.date.today()
        print("*数据来源：雅虎财经，"+str(today))
    
    return VaR_1d,VaR_ratio 

if __name__ == '__main__':
    ticker='BABA'
    shares=1000
    today='2020-7-1'
    alpha=0.99
    pastyears=1
    printout=True
    VaR,Ratio=stock_VaR_historical_1d(ticker,shares,today)
#==============================================================================
def VaR_historical_grouping(position,ret_series,future_days=1,alpha=0.99):
    """
    功能：计算VaR基本算法，历史模拟法，持有多天(基本分组法，无组内收益率波动调整)
    输入参数：当前头寸金额，历史日收益率序列，未来持有天数，置信度
    注意：分组法需要更长时间的历史数据，不然准确度不高！
    输出：持有多天的VaR(金额)
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import pandas as pd
    #将收益率序列转变为df
    r1=pd.DataFrame(r)
    
    #定义组内累计收益率计算方法
    cumret=lambda x:(x+1.0).prod()-1.0
    #使用滚动窗口分组，计算各组收益率gret
    r1['gret']=r1.rolling(future_days).apply(cumret)
    #将各组收益率转换为组收益率序列
    r2=pd.Series(r1['gret'])
    
    #利用单日历史模拟法计算VaR
    VaR_days=VaR_historical_1d(position,r2,alpha)

    return -abs(VaR_days) 

if __name__ == '__main__':
    ticker='BABA'
    start='2020-6-1'
    today='2020-7-1'
    p=get_stock_quotes(ticker,start,today)
    ret_series=get_ret_series(p)    
    future_days=2

    shares=1000
    alpha=0.99
    pastyears=1
    printout=True   
    position=1

    VaR=VaR_historical_grouping(position,ret_series,future_days)
    

#==============================================================================
def VaR_historical(position,ret_series,future_days=1,alpha=0.99):
    """
    功能：同名函数，为了后面函数名称合成方便
    """
    VaR=VaR_historical_grouping(position,ret_series,future_days,alpha)
    return VaR

#==============================================================================
def stock_VaR_historical_grouping(ticker,shares,today, \
                future_days=1,alpha=0.99,pastyears=1,printout=True):
    """
    功能：计算持有股票的VaR，历史模拟法，持有多天(基本分组法，无组内收益率波动调整)
    输入参数：股票代码，持有股数，当前日期，未来持有天数，置信度，使用历史数据的年数
    注意：分组法需要更长时间的历史数据，不然准确度不高！
    建议：pastyears的数值大于等于future_days的数值
    输出：基本分组法，持有多天的VaR(金额和比率)
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)
    if (p is None) or (len(p)==0): 
        print("#Error(stock_VaR_historical_grouping): no obs retrieved.")
        return None,None
    p_end=get_end_price(p); position=shares*p_end
    
    #产生历史收益率序列
    r=get_ret_series(p)
    #分组计算
    VaR_days=VaR_historical_grouping(position,r,future_days,alpha)
    #最大为全损
    if abs(VaR_days) > position: VaR_days=-position
    
    VaR_ratio=abs(VaR_days/position)
    
    if printout == True:
        print("=== 计算在险价值VaR：分组历史模拟方法 ===")
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有日期    :",today)
        print("预计持有天数:",future_days)
        print("置信度      :",alpha*100,'%',sep='')
        print("VaR金额     :",format(round(VaR_days,2),','))
        print("VaR比例     :",round(VaR_ratio*100,2),'%',sep='')      
        
        import datetime as dt; today=dt.date.today()
        print("*数据来源：雅虎财经，"+str(today))
    
    return -abs(VaR_days),VaR_ratio 
  
#==============================================================================
def VaR_montecarlo(position,ret_series, \
                future_days=1,alpha=0.99,random=10000,mctype='random'):
    """
    功能：计算VaR基本算法，蒙特卡洛模拟法，持有多日
    输入参数：当前头寸金额，历史日收益率序列，未来持有天数，置信度，重复模拟次数
    注：重复模拟次数越多，准确率就越高，但耗时也越多
    输出：持有多天的VaR(金额)
    """
    #去掉空值
    r=ret_series[~ret_series.isnull()]
    
    import pandas as pd
    #蒙特卡洛模拟类型：随机数or超采样
    if mctype=='random':    #随机数产生新的序列
        import numpy as np
        #取得历史日收益率的均值和标准差
        miu=np.mean(r)
        sigma=np.std(r)
        #生成随机数序列
        np.random.seed(12345)
        #按照历史日收益率的均值和标准差重复模拟一定次数，生成新的日收益率序列
        RR=pd.Series(np.random.normal(miu,sigma,random))
    else:   #超采样产生新的序列
        #将收益率序列转变为df
        r1=pd.DataFrame(r)
        r2=r1.sample(n=random,replace=True)
        r2.sort_index(inplace=True)
        RR=pd.Series(r2.iloc[:,0])
        
    #基于新的日收益率序列，使用标准正态法计算VaR
    VaR_days=VaR_historical_grouping(position,RR,future_days,alpha)
    
    return -abs(VaR_days)


#==============================================================================
def stock_VaR_montecarlo(ticker,shares,today,future_days=1,alpha=0.99, \
                         pastyears=1,random=10000,printout=True,mctype='random'):
    """
    功能：计算持有股票的VaR，蒙特卡洛模拟法，持有多日
    输入参数：股票代码，持有股数，当前日期，未来持有天数，置信度，使用历史数据的年数，
    重复模拟次数
    输出：持有股票多天的VaR(金额和比率)
    """
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if (p is None) or (len(p)==0): 
        print("#Error(stock_VaR_montecarlo): no obs retrieved.")               
        return None,None
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end
    
    VaR_days=VaR_montecarlo(position,r,future_days,alpha,random,mctype=mctype)
    ratio=abs(VaR_days/position)
    #最大为全损
    if abs(VaR_days) > position: VaR_days=-position
    
    if printout == True:
        print("= 计算在险价值VaR：蒙特卡洛模拟法 =")
        print("持有日期    :",today)
        print("持有股票    :",codetranslate(ticker))
        print("持有股数    :",format(shares,','))
        print("持有头寸    :",format(round(position,2),','))
        print("预计持有天数:",future_days)
        print("置信度      : ",alpha*100,'%',sep='')
        print("序列生成方法:",mctype)
        print("VaR金额     :",format(round(VaR_days,2),','))
        #四舍五不入
        print("VaR比例     : ",(int(ratio*10000))/100,'%',sep='')      
        
        import datetime as dt; today=dt.date.today()
        print("*数据来源：雅虎财经，"+str(today))
    
    return VaR_days,ratio


#==============================================================================
def calc_VaR_tlcp(ticker,today,alpha=0.99, \
                  pastyears=1,model="normal_standard",printout=True):
    """
    功能：计算一只股票的VaR全损临界点，即VaR达到持有全部头寸的最小天数
    输入：股票代码，持有日期，置信度，使用历史数据的年数，
    使用的模型，是否打印结果
    输出：股票的VaR全损临界点天数，与使用的模型有关
    显示：无
    """
    print("\n...Calculatiing tlcp of",ticker,"\b, please wait...")
    
    #检查model类型
    modellist=['normal_standard','normal_modified','montecarlo']
    model=model.lower()
    if model not in modellist:
        print("#Error(calc_VaR_tlcp): Unsupported type of model")
        print("Information:",model)
        print("Supported models:",modellist)
        return    
    
    #抓取股价，计算历史收益率序列和当前头寸
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if (p is None) or (len(p)==0): 
        print("#Error(calc_VaR_tlcp): no obs retrieved.")
        return None
    r=get_ret_series(p)
    p_end=get_end_price(p)
    shares=1
    position=shares*p_end
    
    #第1轮搜索，步长1000天
    func="VaR_"+model
    import numpy as np
    stop1=0; max1=9000
    for d in np.arange(0,max1+1000,1000):
        VaR=eval(func)(position,r,d,alpha)
        ratio=abs(VaR)/position*100
        if ratio >= 100.0:
            stop1=d
            break
    if stop1==0:
        if printout == True:
            print("=== VaR的全损临界点TLCP ===")
            print("持有股票  :",codetranslate(ticker))
            print("持有日期  :",today)
            print("使用的模型:",model)
            print("置信度    : ",alpha*100,'%',sep='')
            print("TLCP天数  : >",format(max1,','))
            
            print("*注：实际发生全损的概率极小")
            import datetime as dt; today=dt.date.today()
            print("*数据来源：雅虎财经，"+str(today))
        return max1        
    
    #第2轮搜索，步长100天  
    stop2=100
    for d in np.arange(stop1-1000,stop1+100,100):
        VaR=eval(func)(position,r,d,alpha)
        ratio=abs(VaR)/position*100
        if ratio >= 100.0:
            stop2=d
            break        
    
    #第3轮搜索，步长10天    
    stop3=10
    for d in np.arange(stop2-100,stop2+10,10):
        VaR=eval(func)(position,r,d,alpha)
        ratio=abs(VaR)/position*100
        if ratio >= 100.0:
            stop3=d
            break 
    
    #第4轮搜索，步长1天  
    stop4=1
    for d in np.arange(stop3-10,stop3+1,1):
        VaR=eval(func)(position,r,d,alpha)
        ratio=abs(VaR)/position*100
        if ratio >= 100.0:
            stop4=d
            break 
    
    if printout == True:
        print("=== VaR的全损临界点TLCP ===")
        print("持有股票  :",codetranslate(ticker))
        print("持有日期  :",today)
        print("使用的模型:",model)
        print("置信度    : ",alpha*100,'%',sep='')
        print("TLCP天数  : >",format(stop4,','))

    return stop4

if __name__ == "__main__":
    ticker='AAPL'
    today='2020-7-20'
    alpha=0.99
    pastyears=1
    model="montecarlo"
    printout=True
    tlcp=calc_VaR_tlcp('MSFT','2019-8-8')

#==============================================================================
def series_VaR_tlcp(tickerlist,today,alpha=0.99,pastyears=1,model="montecarlo"):
    """
    功能：计算股票列表中各个股票的VaR全损临界点(天数)
    输出：各个股票的VaR全损临界点(天数)列表
    显示：柱状图，从小到大排序
    """

    #仅测试用
    #tickerlist=['BABA','JD','VIPS','PDD']
    #today='2019-8-8'
    #alpha=0.99
    #pastyears=1
    #model="montecarlo"
    print("\n*** It may need very long time to calculate, please wait ...")

    import pandas as pd
    tlcpdf=pd.DataFrame(columns=['Ticker','Holding date','Alpha','Model','TLCP'])
    for t in tickerlist:
        tlcp=calc_VaR_tlcp(t,today,alpha,pastyears,model,printout=False)
        row=pd.Series({'Ticker':t,'Holding date':today,'Alpha':alpha, \
                       'Model':model,'TLCP':tlcp})
        tlcpdf=tlcpdf.append(row,ignore_index=True)   
        #print("TLCP for stock ",t,": ",tlcp," days",sep='')
    #tlcpdf=tlcpdf.set_index('Ticker')    
    tlcpdf=tlcpdf.sort_values(axis=0,ascending=True,by = 'TLCP')
    #tlcpdf.reset_index(drop=True)    
    tlcpdf.set_index('Ticker',drop=True,inplace=True)
    """
    #ax=plt.figure(figsize=(8,4))
    tlcpdf.plot.barh(x='Ticker',y='TLCP',color='r',grid=True)
    plt.ylabel('股票')
    titlel1="VaR全损临界点(TLCP)"
    titlel2="\n"+str(today)+", 使用模型"+model+", 置信度"+str(alpha*100)+"%"
    titletxt=titlel1+titlel2
    plt.title(titletxt)
    
    import datetime as dt; today=dt.date.today() 
    footnote='TLCP天数-->'+"\n数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)
    plt.show() 
    """
    titlel1="VaR全损临界点(TLCP)"
    titlel2="\n"+str(today)+", 使用模型"+model+", 置信度"+str(alpha*100)+"%"
    titletxt=titlel1+titlel2    
    import datetime as dt; today=dt.date.today() 
    footnote='TLCP天数-->'+"\n数据来源：雅虎财经，"+str(today)    
    import siat.grafix as g
    g.plot_barh(tlcpdf,'TLCP',titletxt,footnote)
    
    print("===== VaR全损临界点(TLCP, 天数) =====")
    print("持有日期  :",today)
    print("使用的模型:",model)
    print("置信度    :",alpha*100,'%',sep='') 
    print(tlcpdf)
    print("*数据来源：雅虎财经，"+str(today))
    
    return tlcpdf


#==============================================================================
def plot_VaR_days_changes(ticker,shares,today,dayslist,alpha=0.99, \
                        pastyears=1,model="montecarlo", \
                        printout=True, markertype='.'):
    """
    功能：计算持有股票的VaR，持有多个日期
    输入参数：股票代码，持有股数，当前日期，未来持有天数列表，置信度，
    使用历史数据的年数，算法，默认为蒙特卡洛模拟法
    输出：持有股票多个日期的VaR(金额和比率)，数据框格式
    """
    
    #仅测试用
    #ticker='AAPL'
    #shares=1000
    #today='2019-8-8'
    #dayslist=[1,5,10,15,30,90,180]
    #alpha=0.99
    #pastyears=1
    #model="montecarlo"
    #printout=True    
    
    print("\n... Calculating VaR may take great time, please wait...")
    #抓取股价，计算历史收益率序列和当前头寸
    start=get_start_date(today,pastyears)
    p=get_stock_quotes(ticker,start,today)  
    if (p is None) or (len(p)==0): return None
    r=get_ret_series(p)
    p_end=get_end_price(p)
    position=shares*p_end

    #计算各个持有天数的VaR
    import pandas as pd
    vardf=pd.DataFrame(columns=['ticker','holding date','holding days','VaR','ratio%'])
    for d in dayslist:
        func="VaR_"+model
        #print(ticker,shares,today,d,alpha)
        VaR=round((eval(func)(position,r,d,alpha)),0)
        ratio=round((abs(VaR)/position*100),2)
        if ratio >= 100.0: ratio=100.0
        s = pd.Series({'ticker':ticker,'holding date':today,'holding day(s)':d, \
                       'VaR':VaR,'ratio%':ratio})
        vardf=vardf.append(s,ignore_index=True)
    vardf=vardf.set_index(['holding day(s)']) 

    #绘图
    if printout == True:
        plt.figure(figsize=(8,4))
        #绘制曲线：ratio
        plt.plot(vardf['ratio%'],color='red',lw=2,label=codetranslate(ticker),marker=markertype)
        #设置标题、图例、坐标轴标签
        plt.ylabel('VaR比例%')
        plt.legend(loc='best')
        titlel1="持有股票期间与VaR比例的变化趋势"
        titlel2="\n"+codetranslate(ticker)+", "+today+", 使用模型"+model
        titletxt=titlel1+titlel2
        plt.title(titletxt)
        
        import datetime as dt; today=dt.date.today() 
        footnote='持有天数-->'+"\n数据来源：雅虎财经，"+str(today)
        plt.xlabel(footnote)
        plt.show()

    return vardf  

#==============================================================================
def compare_VaR_days_changes(tickerlist,shares,today,dayslist,alpha=0.99, \
                          pastyears=1,model="montecarlo"):
    """
    功能：计算多个股票的持有VaR，分别持有多个日期
    输入参数：股票代码列表，持有股数，当前日期，未来持有天数列表，置信度，
    使用历史数据的年数，算法，默认为蒙特卡洛模拟法
    输出：比较多支股票多个日期的VaR(金额和比率)，数据框格式
    显示：绘图比较
    """
    
    #仅测试用
    #tickerlist=['BABA','JD','PDD']
    #shares=1000
    #today='2019-8-8'
    #dayslist=[1,5,10,15,30,60,90,120,150,180,210,240,270,300,330,365]
    #alpha=0.99
    #pastyears=1
    #model="montecarlo"

    #检查model类型
    modellist=['normal_standard','normal_modified','historical','montecarlo']
    model=model.lower()
    if model not in modellist:
        print("#Error(compare_VaR_days_changes): Unsupported type of model")
        print("Information:",model)
        print("Supported models:",modellist)
        return
    
    #依次绘图tickerlist中的各个股票
    plt.figure(figsize=(8,4))
    
    markerlist=['.','o','s','*','+','x','1','2']
    for t in tickerlist:
        pos=tickerlist.index(t)
        thismarker=markerlist[pos]
        vardf=plot_VaR_days_changes(t,shares,today,dayslist,alpha, \
                pastyears,model,printout=False,markertype=thismarker)
        if vardf is None: 
            print("#Error(compare_VaR_days_changes): No available data")
            print("Information:",t,today,pastyears,model)
            continue        
        
        plt.plot(vardf['ratio%'],lw=2,label=codetranslate(t),marker=thismarker)
    
    #设置标题、图例、坐标轴标签
    plt.ylabel('VaR比例%')
    plt.legend(loc='best')
    titlel1="持有股票期间与VaR比例的变化趋势"
    titlel2="\n股票: "+str(tickerlist)+", "+today+", 使用模型"+model
    titlel3="\n注: 水平线表示VaR全损"
    titletxt=titlel1+titlel2+titlel3
    plt.title(titletxt)
    
    import datetime as dt; today=dt.date.today() 
    footnote='持有天数-->'+"\n数据来源：雅虎财经，"+str(today)
    plt.xlabel(footnote)
    plt.show()

    return

if __name__ =="__main__":
    dayslist=[1,5,15,30,60,90,180,365]
    compare_VaR_days_changes('AAPL','JD',1000,'2019-8-8',dayslist)

#==============================================================================
def get_VaR_allmodels(ticker,shares,today, \
                      future_days=1,alpha=0.99,pastyears=1):
    """
    功能：一次性计算各种模型方法的VaR
    输入参数：股票代码，持有股数，当前日期，未来持有天数，置信度，使用历史数据的年数
    股票代码：美股代码格式
    当前日期：datetime类型
    输出参数：各种模型的VaR金额和比率
    模型：标准正态法，修正正态法，历史模拟基本分组法，蒙特卡洛模拟法
    """
    #计算最低年数以保证历史模拟法至少拥有30个分组来确保结果的有效性
    py=int(future_days*30/252)+1
    if py > pastyears:
        print("Warning: Historical samples too few, historical simulation may not work properly")
    
    #抓取开始日期
    start=get_start_date(today,pastyears)
    #抓取股价序列
    quotes=get_stock_quotes(ticker,start,today)
    if (quotes is None) or (len(quotes)==0): 
        print("#Error(get_VaR_allmodels): no obs retrieved.")
        return None,None
    #计算当前头寸
    position=shares*get_end_price(quotes)
    #计算收益率序列
    rets=get_ret_series(quotes)
    
    #模型1：标准正太法。加0.5取整相当于四舍五入
    VaR1=int(VaR_normal_standard(position,rets,future_days,alpha)+0.5)
    ratio1=abs(round(VaR1/position,4))
    
    #模型2：修正正太法
    VaR2=int(VaR_normal_modified(position,rets,future_days,alpha)+0.5)
    ratio2=abs(round(VaR2/position,4))
    
    #模型3：历史模拟法-分组法
    #VaR3=int(VaR_historical_cumulative(position,rets,future_days,alpha)+0.5)
    VaR3=int(VaR_historical_grouping(position,rets,future_days,alpha)+0.5)
    ratio3=abs(round(VaR3/position,4))
                
    #模型4：蒙特卡洛模拟法(标准正态分布)
    VaR4=int(VaR_montecarlo(position,rets,future_days,alpha)+0.5)
    ratio4=abs(round(VaR4/position,4))
    
    print("= 比较不同模型下VaR的计算结果 =")
    print("持有日期    :",today)
    print("持有股票    :",codetranslate(ticker))
    print("持有股数    :",format(shares,','))
    print("持有头寸    :",format(round(position,2),','))   
    print("预计持有天数:",future_days)
    print("置信度      : ",alpha*100,'%',sep='')
    
    print("\n不同模型下计算的VaR金额:")
    print("标准正态模型  :",format(VaR1,','))
    print("修正正态模型  :",format(VaR2,','))
    print("历史模拟法    :",format(VaR3,','))
    print("蒙特卡洛模拟法:",format(VaR4,','))
    
    print("\n不同模型下计算的VaR比例:")
    print("标准正态模型  :",round(ratio1*100,2),'%',sep='')
    print("修正正态模型  :",round(ratio2*100,2),'%',sep='')
    print("历史模拟法    :",round(ratio3*100,2),'%',sep='')
    print("蒙特卡洛模拟法:",round(ratio4*100,2),'%',sep='')   

    import datetime as dt; today=dt.date.today()
    print("*数据来源：雅虎财经，"+str(today))    
    
    VaRlist=[VaR1,VaR2,VaR3,VaR4]
    ratiolist=[ratio1,ratio2,ratio3,ratio4]
    
    return VaRlist,ratiolist 

#==============================================================================
def get_ret_portfolio(tickerlist,shareslist,today,pastyears=1):
    """
    功能：计算投资组合的历史收益率序列和当前头寸金额
    输入参数：投资组合的股票列表，投资组合的持有股数列表，当前日期，使用历史数据的年数
    输出参数：投资组合的历史日收益率序列，当前持有头寸金额
    注意：
    #tickerlist为字符串列表类型，内容为投资组合内各个股票的代码
    #shareslist为整数列表类型，内容为投资组合内各个股票的持股数量，不是金额
    #tickerlist和shareslist中的元素个数必须一一对等
    """
    import pandas as pd
    start=get_start_date(today,pastyears)
    prices=pd.DataFrame()
    for t in range(len(tickerlist)):
        ticker=tickerlist[t]
        q=get_stock_quotes(ticker,start,today)
        q['date']=q.index.strftime('%Y'+'%m'+'%d')
        q['ticker']=ticker
        q['price']=q['Close']    
        q['shares']=shareslist[t]
        q['value']=q['price']*q['shares']
        prices=prices.append(q)

    prices2=prices.sort_values('date')
    groupKey='date'; groupWay=prices2.groupby(groupKey)
    prices3=groupWay['value'].sum()  
    position=prices3[-1]
    ret=prices3.pct_change()
    ret=ret.dropna()
    return ret,position

#==============================================================================     
def get_ret_portfolio2(tickerlist,shareslist,today,pastyears=1):
    """
    功能：计算投资组合的历史收益率序列、当前头寸金额和各个成分股的历史日收益率序列 
    输入参数：投资组合的股票列表，投资组合的持有股数列表，当前日期，使用历史数据的年数
    输出：投资组合的历史日收益率序列，当前持有头寸金额，各个成分股的历史日收益率序列
    注意：
    #tickerlist为字符串列表类型，内容为投资组合内各个股票的代码
    #shareslist为整数列表类型，内容为投资组合内各个股票的持股数量，不是金额
    #tickerlist和shareslist中的元素个数必须一一对等
    """
    import pandas as pd
    start=get_start_date(today,pastyears)
    prices=pd.DataFrame()
    for t in range(len(tickerlist)):
        ticker=tickerlist[t]
        q=get_stock_quotes(ticker,start,today)
        q['date']=q.index.strftime('%Y'+'%m'+'%d')
        q['ticker']=ticker
        q['price']=q['Close']    
        q['shares']=shareslist[t]
        q['value']=q['price']*q['shares']
        q['ret']=q['price'].pct_change()
        prices=prices.append(q)
    
    #portfolio收益率
    prices2=prices.sort_values('date')
    groupKey='date'; groupWay=prices2.groupby(groupKey)
    prices3=groupWay['value'].sum()  
    pret=prices3.pct_change()
    pret=pret.dropna()
    
    #portfolio当前持有头寸
    position=prices3[-1]
    
    #成分股收益率
    sret=prices[['date','ticker','price','shares','value','ret']]
    sret=sret.dropna()
    
    return pret,position,sret

#==============================================================================     
def get_portfolio_info(tickerlist,today,pastyears=1):
    """
    功能：返回一组股票的历史价格和历史日收益率
    输入参数：股票列表，当前日期，使用历史数据的年数
    输出：成分股收益率数据表
    """
    import pandas as pd
    start=get_start_date(today,pastyears)
    prices=pd.DataFrame()

    for t in range(len(tickerlist)):
        ticker=tickerlist[t]
        q=get_stock_quotes(ticker,start,today)
        q['date']=q.index.strftime('%Y'+'%m'+'%d')
        q['ticker']=ticker
        q['price']=q['Close']    
        q['shares']=0
        q['value']=0
        q['ret']=q['price'].pct_change()
        prices=prices.append(q)
   
    #成分股收益率
    tickers_info=prices[['date','ticker','price','shares','value','ret']]
    return tickers_info

#==============================================================================     
def get_portfolio_rets(tickerlist,shareslist,tickers_info,today,pastyears=1):
    """
    功能：计算投资组合的历史日收益率序列和当前头寸金额，不重新抓取股价
    输入参数：股票列表，持有股数列表，股票信息数据框，当前日期，使用历史数据的年数
    输出：投资组合的历史日收益率序列和当前头寸
    注意：
    #tickerlist为字符串列表类型，内容为投资组合内各个股票的代码
    #shareslist为整数列表类型，内容为投资组合内各个股票的持股数量，不是金额
    #tickerlist和shareslist中的元素个数必须一一对等
    """
    import pandas as pd
    start=get_start_date(today,pastyears)
    startdate=start.strftime('%Y'+'%m'+'%d')
    enddate  =today.strftime('%Y'+'%m'+'%d')
    ti=pd.DataFrame()
    for index, row in tickers_info.iterrows():
        if (row['date'] >= startdate) and (row['date'] <= enddate):
            ticker=row['ticker']
            p=tickerlist.index(ticker)
            shares=shareslist[p]
            row['shares']=shares  
            row['value']=row['price']*shares
            ti=ti.append(row,ignore_index=True)
          
    ti1=ti.sort_values('date')
    groupKey='date'; groupWay=ti1.groupby(groupKey)
    ti2=groupWay['value'].sum()  
    position=ti2[-1]
    pret=ti2.pct_change()
    pret=pret.dropna()
    
    return pret,position

#==============================================================================  
def get_ret_cumulative(ret_series):
    """
    功能：计算累计收益率
    输入参数：收益率序列
    输出：累积收益率    
    """
    c=1
    for r in ret_series:
        c=c*(1+r)
    retcum=c-1    
    return retcum

#==============================================================================     
def get_portfolio_ret_risk(tickerlist,shareslist,tickers_info, \
                           today,future_days=1,alpha=0.99,pastyears=1):
    """
    功能：一次性计算投资组合的所有收益率和风险信息，避免重复抓取网络股价
    输入参数：投资组合的股票列表，投资组合的持有股数列表，投资组合的成分股信息，当前日期，
    未来持有天数，置信度，使用历史数据的年数
    输出：投资组合的累积收益率、VaR金额与比率、ES金额与比率；各个成分股的累积收益率、
    VaR金额与比率、ES金额与比率  
    """
    #合成投资组合历史信息
    Pret,Pposition=get_portfolio_rets(tickerlist,shareslist, \
                                tickers_info,today,pastyears)

    #投资组合累积收益率：例13.82%
    Pretcum=round(get_ret_cumulative(Pret),4)
    #投资组合VaR金额和比率：例-$8851，0.0304
    PVaR=int(VaR_normal_standard(Pposition,Pret,future_days,alpha)+0.5)
    PVaRratio=round(abs(PVaR/Pposition),4)
    #投资组合ES金额和比率：例-$10166，0.035
    PES=int(ES_normal_standard(Pposition,Pret,future_days,alpha)+0.5)
    PESratio=round(abs(PES/Pposition),4)

    #成分股累积回报率列表：[-0.0113,0.3278]
    retlist=[]
    for t in tickerlist:
        r=tickers_info[tickers_info['ticker']==t]['ret']
        r=r.dropna()
        retcum=round(get_ret_cumulative(r),4)
        retlist=retlist+[retcum]

    #VaR金额和比率列表：例[-$4650,-$6691],[0.0329,0.0447]
    VaRlist=[]
    VaRratiolist=[]
    for t in range(len(tickerlist)):
        ticker=tickerlist[t]
        r=tickers_info[tickers_info['ticker']==ticker]['ret']
        r=r.dropna()
    
        shares=shareslist[t]
        price=tickers_info[tickers_info['ticker']==ticker]['price'][-1]
        position=shares*price
            
        VaR=int(VaR_normal_standard(position,r,future_days,alpha)+0.5)
        ratio=round(abs(VaR/position),4)
        VaRlist=VaRlist+[VaR]
        VaRratiolist=VaRratiolist+[ratio]

    #ES金额和比率列表：例[-$5329,-$7695],[0.0377,0.0515]
    ESlist=[]
    ESratiolist=[]
    for t in range(len(tickerlist)):
        ticker=tickerlist[t]
        r=tickers_info[tickers_info['ticker']==ticker]['ret']
        r=r.dropna()
            
        shares=shareslist[t]
        price=tickers_info[tickers_info['ticker']==ticker]['price'][-1]
        position=shares*price
            
        ES=int(ES_normal_standard(position,r,future_days,alpha)+0.5)
        ratio=round(abs(ES/position),4)
        ESlist=ESlist+[ES]
        ESratiolist=ESratiolist+[ratio]    
    
    return Pretcum,PVaR,PVaRratio,PES,PESratio, \
            retlist,VaRlist,VaRratiolist,ESlist,ESratiolist


#==============================================================================     
def get_portfolio_ret_risk2(tickerlist,shareslist,tickers_info, \
                           today,future_days=1,alpha=0.99,pastyears=1):
    """
    功能：一次性计算投资组合的所有收益率和风险信息，仅返回比率
    输入参数：投资组合的股票列表，投资组合的持有股数列表，投资组合的成分股信息，当前日期，
    未来持有天数，置信度，使用历史数据的年数
    输出参数：投资组合的累积收益率、VaR比率、ES比率；各个成分股的累积收益率、VaR比率、ES比率  
    """
    Pretcum,PVaR,PVaRratio,PES,PESratio, \
            retlist,VaRlist,VaRratiolist,ESlist,ESratiolist= \
            get_portfolio_ret_risk(tickerlist,shareslist,tickers_info, \
                           today,future_days,alpha,pastyears)

    return Pretcum,PVaRratio,PESratio,retlist,VaRratiolist,ESratiolist 


#==============================================================================     
def get_portfolio_ret_risk3(tickerlist,shareslist,tickers_info, \
                           today,future_days=1,alpha=0.99,pastyears=1):
    """
    功能：一次性计算投资组合的所有收益率和风险信息，仅返回投资组合的比率
    输入参数：投资组合的股票列表，投资组合的持有股数列表，投资组合的成分股信息，当前日期，
    未来持有天数，置信度，使用历史数据的年数
    输出：投资组合的累积收益率、VaR比率、ES比率
    """
    Pretcum,PVaR,PVaRratio,PES,PESratio, \
            retlist,VaRlist,VaRratiolist,ESlist,ESratiolist= \
            get_portfolio_ret_risk(tickerlist,shareslist,tickers_info, \
                           today,future_days,alpha,pastyears)

    return Pretcum,PVaRratio,PESratio   


#==============================================================================








# -*- coding: utf-8 -*-
"""
本模块功能：提供全球股票ESG信息
所属工具包：证券投资分析工具SIAT 
SIAT：Security Investment Analysis Tool
创建日期：2019年8月18日
最新修订日期：2020年7月25日
作者：王德宏 (WANG Dehong, Peter)
作者单位：北京外国语大学国际商学院
版权所有：王德宏
用途限制：仅限研究与教学使用，不可商用！商用需要额外授权。
特别声明：作者不对使用本工具进行证券投资导致的任何损益负责！
"""
#==============================================================================
#屏蔽所有警告性信息
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.translate import *
from siat.grafix import *
#==============================================================================
#==============================================================================
#以下使用yfinance数据源
#==============================================================================
def get_esg(stocklist):
    """
    功能：根据股票代码列表，抓取企业最新的可持续性发展ESG数据
    输入参数：
    stocklist：股票代码列表，例如单个股票["AAPL"], 多只股票["AAPL","MSFT","GOOG"]
    输出参数：    
    企业最新的可持续性发展ESG数据，数据框
    """
        
    #引用插件
    try:
        import yfinance as yf
    except:
        print("#Error(get_esg), need plugin: yfinance")
        print("   Solution: open Anaconda Prompt, type in command \"pip install yfinance\", Enter!")
        return None
    tickerlist=stocklist.copy()
    
    #测试数据，使用后请注释掉
    """
    tickerlist=["PDD","MSFT","BABA","JD","GOOG"]
    """

    #处理股票列表中的第一只股票，跳过无数据的项目
    skiplist=[]
    for t in tickerlist:
        tp=yf.Ticker(t)        
        try:
            print("...正在搜索ESG信息:",t,"...",end='')
            sst=tp.sustainability
            sst.rename(columns={'Value':t},inplace=True)
            sstt=sst.T            
        except: #本项目无数据，进入下一次循环
            print("未找到:-(")
            skiplist=skiplist+[t]
            continue
        skiplist=skiplist+[t]
        print("完成!")
        break

    #仅保留尚未处理的项目
    for t in skiplist: tickerlist.remove(t)
    
    #处理股票列表中的其他股票    
    for t in tickerlist:
        #print("---stock:",t)
        tp=yf.Ticker(t)        
        try: 
            print("...正在搜索ESG信息:",t,"...",end='')
            sst1=tp.sustainability
            sst1.rename(columns={'Value':t},inplace=True)
        except: 
            print("未找到:-(")
            continue #未抓取到数据
        sst1t=sst1.T
        sstt=sstt.append([sst1t])
        print("完成!")
    
    try:
        if len(sstt)==0: return None
    except:
        return None
    
    #只保留需要的列
    sust=sstt[['totalEsg','percentile','esgPerformance','environmentScore', \
               'environmentPercentile','socialScore','socialPercentile', \
               'governanceScore','governancePercentile','peerGroup','peerCount']].copy()
    sust.rename(columns={'totalEsg':'ESGscore','percentile':'ESGpercentile', \
            'esgPerformance':'ESGperformance','environmentScore':'EPscore', \
               'environmentPercentile':'EPpercentile','socialScore':'CSRscore', \
               'socialPercentile':'CSRpercentile','governanceScore':'CGscore', \
               'governancePercentile':'CGpercentile', \
               'peerGroup':'Peer Group','peerCount':'Count'},inplace=True)
    
    return sust

if __name__ =="__main__":
    stocklist=["VIPS","BABA","JD","MSFT","WMT"]
    sust=get_esg(stocklist)


#==============================================================================
def print_esg(sustainability,option="ESG"):
    """
    功能：显示企业的可持续性发展数据
    输入参数：
    sustainability：抓取到的企业可持续性数据框
    输出参数：无    
    """
    
    if not (option in ['ESG','EP','CSR','CG']):
        print("...Error 01(print_sustainability): only ESG/EP/CSR/CG are valid")
        return
    
    import datetime as dt
    today=dt.date.today()
    s=sustainability.copy()
    
    #修改评级为中文
    s['ESGperformance']=s['ESGperformance'].apply(lambda x: "高" if x=='OUT_PERF' else x)
    s['ESGperformance']=s['ESGperformance'].apply(lambda x: "中" if x=='AVG_PERF' else x)
    
    #列改中文名，并删除含0的无效列
    colnames=list(s)
    for c in colnames:
        if len(s[s[c]==0]) >= 1:
            del s[c]; continue
        s.rename(columns={c:ectranslate(c)},inplace=True)
    
    #打印输出对齐
    import pandas as pd
    pd.set_option('display.unicode.ambiguous_as_wide', True)
    pd.set_option('display.unicode.east_asian_width', True)
    pd.set_option('display.width', 180) # 设置打印宽度(**重要**)
    
    #显示分数和分位数
    colnames=list(s)
    printnames=[]
    if option=="ESG":
        s=s.sort_values([ectranslate('ESGscore')],ascending=False)
        esgnames=[ectranslate('ESGscore'),ectranslate('ESGperformance'), \
                  ectranslate('ESGpercentile'),ectranslate('Peer Group')]
        for i in esgnames:
            if i in colnames: printnames=printnames+[i]
        print("\n=== 企业发展可持续性：ESG综合风险 ===\n")
        print(s[printnames])
        print("来源: 雅虎财经,",str(today))

    if option=="EP":
        s=s.sort_values([ectranslate('EPscore')],ascending=False)
        esgnames=[ectranslate('EPscore'), \
                  ectranslate('EPpercentile'),ectranslate('Peer Group')]
        for i in esgnames:
            if i in colnames: printnames=printnames+[i]
        print("\n=== 企业发展可持续性：环保风险指数 ===\n")
        print(s[printnames])
        print("来源: 雅虎财经,",str(today))

    if option=="CSR":
        s=s.sort_values([ectranslate('CSRscore')],ascending=False)
        esgnames=[ectranslate('CSRscore'), \
                  ectranslate('CSRpercentile'),ectranslate('Peer Group')]
        for i in esgnames:
            if i in colnames: printnames=printnames+[i]
        print("\n=== 企业发展可持续性：社会责任风险指数 ===\n")
        print(s[printnames])
        print("来源: 雅虎财经,",str(today))

    if option=="CG":
        s=s.sort_values([ectranslate('CGscore')],ascending=False)
        esgnames=[ectranslate('CGscore'), \
                  ectranslate('CGpercentile'),ectranslate('Peer Group')]
        for i in esgnames:
            if i in colnames: printnames=printnames+[i]
        print("\n=== 企业发展可持续性：公司治理风险指数 ===\n")
        print(s[printnames])
        print("来源: 雅虎财经,",str(today))
    
    return

if __name__ =="__main__":
    print_esg(sust,option="ESG")
    print_esg(sust,option="EP")
    print_esg(sust,option="CSR")
    print_esg(sust,option="CG")
    print_esg(sust,option="ABC")


    
#==============================================================================
def ploth_esg(sustainability,option="ESG"):
    """
    功能：绘制ESG水平直方图
    输入参数：抓取到的企业可持续性数据框sustainability；选项option。
    输出参数：无    
    """
        
    if not (option in ['ESG','EP','CSR','CG']):
        print("...Error 01(ploth_esg): only ESG/EP/CSR/CG are valid")
        return
        
    s=sustainability.copy()
    import datetime as dt; today=dt.date.today()
    footnote="注：数值越小，风险越低。来源：雅虎财经，"+str(today)

    #绘制分数和分位数图     
    if option=="ESG":
        #排序
        s=s.sort_values(['ESGscore'],ascending=True)    
        titletxt1="企业发展可持续性：ESG综合风险指数"
        plot_barh(s,'ESGscore',titletxt1,footnote)
        titletxt2="企业发展可持续性：ESG综合风险行业分位数%"
        plot_barh(s,'ESGpercentile',titletxt2,footnote)

    if option=="EP":
        s=s.sort_values(['EPscore'],ascending=True)        
        titletxt="企业发展可持续性：环保风险指数"
        plot_barh(s,'EPscore',titletxt,footnote)

    if option=="CSR":
        s=s.sort_values(['CSRscore'],ascending=True)        
        titletxt="企业发展可持续性：社会责任风险指数"
        plot_barh(s,'CSRscore',titletxt,footnote) 

    if option=="CG":
        s=s.sort_values(['CGscore'],ascending=True)        
        titletxt="企业发展可持续性：公司治理风险指数"
        plot_barh(s,'CGscore',titletxt,footnote)  
        
    return 

if __name__ =="__main__":
    ploth_esg(sust,option="ESG")
    ploth_esg(sust,option="EP")
    ploth_esg(sust,option="CSR")
    ploth_esg(sust,option="CG")

#==============================================================================
def esg(stocklist):
    """
    功能：抓取、打印和绘图企业的可持续性发展数据，演示用
    输入参数：
    stocklist：股票代码列表，例如单个股票["AAPL"], 多只股票["AAPL","MSFT","GOOG"]
    输出参数：    
    企业最新的可持续性发展数据，数据框    
    """
        
    #抓取数据
    try:
        sust=get_esg(stocklist)
    except:
        print("#Error(esg), fail to get ESG data for",stocklist)
        return None
    if sust is None:
        print("#Error(esg), fail to get ESG data for",stocklist)
        return None
        
    #处理小数点
    from pandas.api.types import is_numeric_dtype
    cols=list(sust)    
    for c in cols:
        if is_numeric_dtype(sust[c]):
            sust[c]=round(sust[c],2)        
    
    #打印和绘图ESG
    ploth_esg(sust,option="ESG")
    print_esg(sust,option="ESG")
    #打印和绘图EP
    ploth_esg(sust,option="EP")
    print_esg(sust,option="EP")
    #打印和绘图CSR
    ploth_esg(sust,option="CSR")
    print_esg(sust,option="CSR")
    #打印和绘图CG
    ploth_esg(sust,option="CG")
    print_esg(sust,option="CG")

    return sust

if __name__ =="__main__":
    stocklist1=["AMZN","EBAY","BABA"]
    sust1=esg(stocklist1)
    stocklist2=["AMZN","EBAY","BABA","JD","VIPS","WMT"]
    sust2=esg(stocklist2)    

#==============================================================================
def portfolio_esg(portfolio):
    """
    功能：抓取、打印和绘图投资组合portfolio的可持续性发展数据，演示用
    输入参数：
    企业最新的可持续性发展数据，数据框    
    """
    #解构投资组合
    _,_,stocklist,_=decompose_portfolio(portfolio)
    
    #抓取数据
    try:
        sust=get_esg(stocklist)
    except:
        print("#Error(portfolio_esg), fail to get ESG data for",stocklist)
        return None
    if sust is None:
        #print("#Error(portfolio_esg), fail to get ESG data for",stocklist)
        return None
        
    #处理小数点
    from pandas.api.types import is_numeric_dtype
    cols=list(sust)    
    for c in cols:
        if is_numeric_dtype(sust[c]):
            sust[c]=round(sust[c],2)        
            
    #显示结果
    print("\n***** Portfolio ESG Risk *****")
    print("Portfolio:",stocklist)
    #显示各个成分股的ESG分数
    sust['Stock']=sust.index
    esgdf=sust[['Stock','ESGscore','EPscore','CSRscore','CGscore']]
    print(esgdf.to_string(index=False))
    
    print("\nPortfolio ESG Evaluation:")
    #木桶短板：EPScore
    esg_ep=esgdf.sort_values(['EPscore'], ascending = True)
    p_ep=esg_ep['EPscore'][-1]
    p_ep_stock=esg_ep.index[-1]   
    str_ep="   EP score (from "+str(p_ep_stock)+")"
    len_ep=len(str_ep)

    #木桶短板：CSRScore
    esg_csr=esgdf.sort_values(['CSRscore'], ascending = True)
    p_csr=esg_csr['CSRscore'][-1]
    p_csr_stock=esg_csr.index[-1] 
    str_csr="   CSR score (from "+str(p_csr_stock)+")"
    len_csr=len(str_csr)
    
    #木桶短板：CGScore
    esg_cg=esgdf.sort_values(['CGscore'], ascending = True)
    p_cg=esg_cg['CGscore'][-1]
    p_cg_stock=esg_cg.index[-1]     
    str_cg="   CG score (from "+str(p_cg_stock)+")"
    len_cg=len(str_cg)

    str_esg="   Overall ESG score"
    len_esg=len(str_esg)
    
    #计算对齐冒号中间需要的空格数目
    len_max=max(len_ep,len_csr,len_cg,len_esg)
    str_ep=str_ep+' '*(len_max-len_ep+1)+':'
    str_csr=str_csr+' '*(len_max-len_csr+1)+':'
    str_cg=str_cg+' '*(len_max-len_cg+1)+':'
    str_esg=str_esg+' '*(len_max-len_esg+1)+':'
    
    #对齐打印
    print(str_ep,p_ep)
    print(str_csr,p_csr)    
    print(str_cg,p_cg)      
    #计算投资组合的ESG综合风险
    p_esg=round(p_ep+p_csr+p_cg,2)
    print(str_esg,p_esg)

    import datetime as dt; today=dt.date.today()
    footnote="The higher the score, the higher the risk. \
        \nSource: Yahoo Finance, "+str(today)
    print(footnote)
    
    return p_esg

if __name__ =="__main__":
    #market={'Market':('China','^HSI')}
    market={'Market':('US','^GSPC')}
    #stocks={'0939.HK':2,'1398.HK':1,'3988.HK':3}
    stocks={'VIPS':3,'JD':2,'BABA':1}
    portfolio=dict(market,**stocks)
    esg=portfolio_esg(portfolio)
#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
#====以下使用yahooquery数据源===================================================
if __name__ =="__main__":
    stocklist=["BAC", "TD","PNC"]
    
def get_esg2(stocklist):
    """
    功能：根据股票代码列表，抓取企业最新的可持续性发展ESG数据
    输入参数：
    stocklist：股票代码列表，例如单个股票["AAPL"], 多只股票["AAPL","MSFT","GOOG"]
    输出参数：    
    企业最新的可持续性发展ESG数据，数据框
    """
    
    import pandas as pd
    collist=['symbol','totalEsg','environmentScore','socialScore','governanceScore']
    sust=pd.DataFrame(columns=collist)
    for t in stocklist:
        try:
            info=stock_info(t).T
        except:
            print("#Error(get_esg2): esg info not available for",t)
            continue
        if (info is None) or (len(info)==0):
            print("#Error(get_esg2): failed to get esg info for",t)
            continue
        sub=info[collist]
        sust=pd.concat([sust,sub])
    
    newcols=['Stock','ESGscore','EPscore','CSRscore','CGscore']
    sust.columns=newcols
    """
    sust=sust.rename(columns={'symbol':'Stock','totalEsg':'ESGscore', \
                         'environmentScore':'EPscore', \
                             'socialScore':'CSRscore', \
                                 'governanceScore':'CGscore'})
    """
    sust.set_index('Stock',inplace=True)
    
    return sust

if __name__ =="__main__":
    stocklist=["VIPS","BABA","JD","MSFT","WMT"]
    sust=get_esg(stocklist)

#==============================================================================
if __name__ =="__main__":
    market={'Market':('China','^HSI')}
    stocks={'0700.HK':3,'9618.HK':2,'9988.HK':1}
    portfolio=dict(market,**stocks)

def portfolio_esg2(portfolio):
    """
    功能：抓取、打印和绘图投资组合portfolio的可持续性发展数据，演示用
    输入参数：
    企业最新的可持续性发展数据，数据框    
    """
    #解构投资组合
    _,_,stocklist,_=decompose_portfolio(portfolio)
    
    #抓取数据
    try:
        sust=get_esg2(stocklist)
    except:
        print("#Error(portfolio_esg), fail to get ESG data for",stocklist)
        return None
    if sust is None:
        #print("#Error(portfolio_esg), fail to get ESG data for",stocklist)
        return None
        
    #处理小数点
    from pandas.api.types import is_numeric_dtype
    cols=list(sust)    
    for c in cols:
        if is_numeric_dtype(sust[c]):
            sust[c]=round(sust[c],2)        
            
    #显示结果
    print("\n***** Portfolio ESG Risk *****")
    print("Portfolio:",stocklist)
    #显示各个成分股的ESG分数
    sust['Stock']=sust.index
    esgdf=sust[['Stock','ESGscore','EPscore','CSRscore','CGscore']]
    print(esgdf.to_string(index=False))
    
    print("\nPortfolio ESG Evaluation:")
    #木桶短板：EPScore
    esg_ep=esgdf.sort_values(['EPscore'], ascending = True)
    p_ep=esg_ep['EPscore'][-1]
    p_ep_stock=esg_ep.index[-1]   
    str_ep="   EP score (from "+str(p_ep_stock)+")"
    len_ep=len(str_ep)

    #木桶短板：CSRScore
    esg_csr=esgdf.sort_values(['CSRscore'], ascending = True)
    p_csr=esg_csr['CSRscore'][-1]
    p_csr_stock=esg_csr.index[-1] 
    str_csr="   CSR score (from "+str(p_csr_stock)+")"
    len_csr=len(str_csr)
    
    #木桶短板：CGScore
    esg_cg=esgdf.sort_values(['CGscore'], ascending = True)
    p_cg=esg_cg['CGscore'][-1]
    p_cg_stock=esg_cg.index[-1]     
    str_cg="   CG score (from "+str(p_cg_stock)+")"
    len_cg=len(str_cg)

    str_esg="   Overall ESG score"
    len_esg=len(str_esg)
    
    #计算对齐冒号中间需要的空格数目
    len_max=max(len_ep,len_csr,len_cg,len_esg)
    str_ep=str_ep+' '*(len_max-len_ep+1)+':'
    str_csr=str_csr+' '*(len_max-len_csr+1)+':'
    str_cg=str_cg+' '*(len_max-len_cg+1)+':'
    str_esg=str_esg+' '*(len_max-len_esg+1)+':'
    
    #对齐打印
    print(str_ep,p_ep)
    print(str_csr,p_csr)    
    print(str_cg,p_cg)      
    #计算投资组合的ESG综合风险
    p_esg=round(p_ep+p_csr+p_cg,2)
    print(str_esg,p_esg)

    import datetime as dt; today=dt.date.today()
    footnote="The higher the score, the higher the risk. \
        \nSource: Yahoo Finance, "+str(today)
    print(footnote)
    
    return p_esg

if __name__ =="__main__":
    #market={'Market':('China','^HSI')}
    market={'Market':('US','^GSPC')}
    #stocks={'0939.HK':2,'1398.HK':1,'3988.HK':3}
    stocks={'VIPS':3,'JD':2,'BABA':1}
    portfolio=dict(market,**stocks)
    esg=portfolio_esg(portfolio)
#==============================================================================

#==============================================================================










        


# -*- coding: utf-8 -*-

import os; os.chdir("S:/siat/siat")

import akshare as ak
from siat.common import *

zs=ak.index_investing_global_country_name_url("中国")
df = ak.index_investing_global(country="中国", index_name="上证指数", period="每日", start_date="2021-01-01", end_date="2021-06-05")
print(index_investing_global_df)

df=ak.stock_us_daily(symbol='^DJI', adjust="")


#==============================================================================
if __name__=='__main__':
    symbol='AAPL'
    fromdate='2020-12-1'
    todate='2021-1-31'

def get_price_ak_us(symbol, fromdate, todate, adjust=""):
    """
    抓取单个美股股价
    """
    
    #检查日期期间
    result,start,end=check_period(fromdate,todate)
    if not result:
        print("  #Error(get_price_ak_us): invalid date period from",fromdate,'to',todate)
        return None  
    
    import akshare as ak
    print("  Searching info in Sina Finance for",symbol,"... ...")
    try:
        df=ak.stock_us_daily(symbol=symbol, adjust="")
    except:
        print("  #Warning(get_price_ak_us): no info found for",symbol)
        return None
    else:
        print("  Successfully retrieved",len(df),"records for",symbol)
    
    #选取时间范围
    df1=df[df.index >=start]    
    df2=df1[df1.index <=end] 
    
    df2.rename(columns={'open':'Open','high':'High','low':'Low','close':'Close','volume':'Volume'},inplace=True)
    df2['Ticker']=symbol
    
    return df2

#==============================================================================
if __name__=='__main__':
    symbol='0700.HK'
    fromdate='2020-12-1'
    todate='2021-1-31'

def get_price_ak_hk(symbol, fromdate, todate, adjust=""):
    """
    抓取单个港股股价
    """
    
    #检查日期期间
    result,start,end=check_period(fromdate,todate)
    if not result:
        print("  #Error(get_price_ak_hk): invalid date period from",fromdate,'to',todate)
        return None  
    
    import akshare as ak
    print("  Searching info in Sina Finance for",symbol,"... ...")
    symbol1 = symbol.strip('.HK')
    symbol2 = symbol1.strip('.hk')
    symbol3='0'+symbol2
    try:
        df=ak.stock_hk_daily(symbol=symbol3, adjust="")
    except:
        print("  #Warning(get_price_ak_hk): no info found for",symbol)
        return None
    else:
        print("  Successfully retrieved",len(df),"records for",symbol)
    
    #选取时间范围
    df1=df[df.index >=start]    
    df2=df1[df1.index <=end] 
    
    df2.rename(columns={'open':'Open','high':'High','low':'Low','close':'Close','volume':'Volume'},inplace=True)
    df2['Ticker']=symbol
    
    return df2

if __name__=='__main__':
    df=get_price_ak_hk(symbol, fromdate, todate)
#==============================================================================
def str_replace(str1):
    """
    删除给定字符串中的子串
    """
    replist=['Ltd.','Ltd','Co.','LTD.','CO.',' CO','LTD','Inc.','INC.', \
             'CORPORATION','Corporation','LIMITED','Limited','Company', \
             'COMPANY','(GROUP)','Corp.','CORP','GROUP','Group']
    
    for rc in replist:
        str2=str1.replace(rc, '')
        str1=str2
    
    twlist=[' ',',','，']    
    for tw in twlist:
        str2 = str2.strip(tw)
    
    return str2
    
#==============================================================================
def get_all_stock_names():
    """
    获得股票代码和名称：中国A股、港股、美股
    """
    import akshare as ak
    import pandas as pd
    
    #上证A股
    df_ss=ak.stock_info_sh_name_code()
    df_ss.rename(columns={'COMPANY_ABBR':'CNAME','ENGLISH_ABBR':'ENAME','LISTING_DATE':'LISTING'},inplace=True)
    df_ss['SYMBOL']=df_ss['COMPANY_CODE']+'.SS'
    df_ss_1=df_ss[['SYMBOL','CNAME','ENAME','LISTING']]    
    
    #深证A股
    df_sz=ak.stock_info_sz_name_code(indicator="A股列表")
    df_sz['SYMBOL']=df_sz['A股代码']+'.SZ'
    df_sz.rename(columns={'A股简称':'CNAME','英文名称':'ENAME','A股上市日期':'LISTING'},inplace=True)
    df_sz_1=df_sz[['SYMBOL','CNAME','ENAME','LISTING']]    
    
    #美股
    df_us=ak.get_us_stock_name()
    df_us['LISTING']=' '
    df_us.rename(columns={'symbol':'SYMBOL','name':'ENAME','cname':'CNAME'},inplace=True)
    df_us_1=df_us[['SYMBOL','CNAME','ENAME','LISTING']] 
    
    #港股
    df_hk=ak.stock_hk_spot()
    df_hk['LISTING']=' '
    last4digits=lambda x:x[1:5]
    df_hk['symbol1']=df_hk['symbol'].apply(last4digits)
    df_hk['SYMBOL']=df_hk['symbol1']+'.HK'
    df_hk.rename(columns={'name':'CNAME','engname':'ENAME'},inplace=True)
    df_hk_1=df_hk[['SYMBOL','CNAME','ENAME','LISTING']] 
    
    #合成
    df=pd.concat([df_ss_1,df_sz_1,df_us_1,df_hk_1])
    df.sort_values(by=['SYMBOL'], ascending=True, inplace=True )
    df.reset_index(drop=True,inplace=True)
    
    rep=lambda x:str_replace(x)
    df['CNAME']=df['CNAME'].apply(rep)
    df['ENAME']=df['ENAME'].apply(rep)
    
    #保存:应保存在文件夹S:/siat/siat中，重新生成siat轮子
    df.to_pickle('stock_info.pickle')

    """
    #读出文件
    with open('stock_info.pickle','rb') as test:
        df = pickle.load(test)
    """
    
    return df
#==============================================================================
if __name__=='__main__':
    symbol='0700.HK'
def get_names(symbol):
    """
    从文件中查询证券代码的短名称
    """
    import pickle
    with open('stock_info.pickle','rb') as test:
        df = pickle.load(test)  
        
    df1=df[df['SYMBOL']==symbol]
    if len(df1)==0:
        return None
    else:
        name=df1['CNAME'].values[0]   
    
    return name
#==============================================================================
# 股票预测插件
from stocker import Stocker
amazon=Stocker("AMZN")
amazon.plot_stock()
model,model_data=amazon.create_prophet_model(days=90)
amazon.evaluate_prediction()



# -*- coding: utf-8 -*-
"""
版权：王德宏，北京外国语大学国际商学院
功能：Fama-French股票市场资产定价因子（不包括单独的中国大陆市场）
版本：2.1，2019-10-10
"""
#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
from siat.common import *
from siat.translate import *
from siat.security_prices import *
#==============================================================================
import matplotlib.pyplot as plt

#处理绘图汉字乱码问题
import sys; czxt=sys.platform
if czxt in ['win32','win64']:
    plt.rcParams['font.sans-serif'] = ['SimHei']  # 设置默认字体
    mpfrc={'font.family': 'SimHei'}

if czxt in ['darwin']: #MacOSX
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family': 'Heiti TC'}

if czxt in ['linux']: #website Jupyter
    plt.rcParams['font.family']= ['Heiti TC']
    mpfrc={'font.family':'Heiti TC'}

# 解决保存图像时'-'显示为方块的问题
plt.rcParams['axes.unicode_minus'] = False 
#==============================================================================
def get_ff_factors(start,end,scope='US',factor='FF3',freq='yearly'):
   """
   【支持的因子种类(factor)】
   FF3: FF3各个因子
   FF5: FF5各个因子
   Mom: 动量因子
   ST_Rev: 短期反转因子
   LT_Rev: 长期反转因子

   【支持的国家/地区(scope)】
   US: 美国
   North_America：北美(美国+加拿大)
   Global: 全球
   Global_ex_US: 全球(除美国外)
   Asia_Pacific_ex_Japan: 亚太(除日本外)，拟作近似中国
   Japan: 日本
   Europe: 欧洲

   【支持的取样频率(freq)】
   yearly: 年
   monthly: 月
   weekly: 周(仅支持美国FF3)
   daily: 日
   """
    
   import pandas as pd
   s=pd.DataFrame([
        ['US','FF3','monthly','F-F_Research_Data_Factors',0],
        ['US','FF3','yearly','F-F_Research_Data_Factors',1],
        ['US','FF3','weekly','F-F_Research_Data_Factors_weekly',0],
        ['US','FF3','daily','F-F_Research_Data_Factors_daily',0],
        ['US','FF5','monthly','F-F_Research_Data_5_Factors_2x3',0],
        ['US','FF5','yearly','F-F_Research_Data_5_Factors_2x3',1],
        ['US','FF5','daily','F-F_Research_Data_5_Factors_2x3_daily',0],  
        ['US','Mom','monthly','F-F_Momentum_Factor',0],
        ['US','Mom','yearly','F-F_Momentum_Factor',1],
        ['US','Mom','daily','F-F_Momentum_Factor_daily',0],  
        ['US','ST_Rev','monthly','F-F_ST_Reversal_Factor',0],
        ['US','ST_Rev','yearly','F-F_ST_Reversal_Factor',1],
        ['US','ST_Rev','daily','F-F_ST_Reversal_Factor_daily',0],    
        ['US','LT_Rev','monthly','F-F_LT_Reversal_Factor',0],
        ['US','LT_Rev','yearly','F-F_LT_Reversal_Factor',1],
        ['US','LT_Rev','daily','F-F_LT_Reversal_Factor_daily',0],   \
        
        ['Global','FF3','monthly','Global_3_Factors',0],
        ['Global','FF3','yearly','Global_3_Factors',1],
        ['Global','FF3','daily','Global_3_Factors_Daily',0],   
        ['Global_ex_US','FF3','monthly','Global_ex_US_3_Factors',0],
        ['Global_ex_US','FF3','yearly','Global_ex_US_3_Factors',1],
        ['Global_ex_US','FF3','daily','Global_ex_US_3_Factors_Daily',0],  
        ['Europe','FF3','monthly','Europe_3_Factors',0],
        ['Europe','FF3','yearly','Europe_3_Factors',1],
        ['Europe','FF3','daily','Europe_3_Factors_Daily',0],  
        ['Japan','FF3','monthly','Japan_3_Factors',0],
        ['Japan','FF3','yearly','Japan_3_Factors',1],
        ['Japan','FF3','daily','Japan_3_Factors_Daily',0],    
        ['Asia_Pacific_ex_Japan','FF3','monthly','Asia_Pacific_ex_Japan_3_Factors',0],
        ['Asia_Pacific_ex_Japan','FF3','yearly','Asia_Pacific_ex_Japan_3_Factors',1],
        ['Asia_Pacific_ex_Japan','FF3','daily','Asia_Pacific_ex_Japan_3_Factors_Daily',0],   
        ['North_America','FF3','monthly','North_America_3_Factors',0],
        ['North_America','FF3','yearly','North_America_3_Factors',1],
        ['North_America','FF3','daily','North_America_3_Factors_Daily',0], \
         
        ['Global','FF5','monthly','Global_5_Factors',0],
        ['Global','FF5','yearly','Global_5_Factors',1],
        ['Global','FF5','daily','Global_5_Factors_Daily',0],   
        ['Global_ex_US','FF5','monthly','Global_ex_US_5_Factors',0],
        ['Global_ex_US','FF5','yearly','Global_ex_US_5_Factors',1],
        ['Global_ex_US','FF5','daily','Global_ex_US_5_Factors_Daily',0],  
        ['Europe','FF5','monthly','Europe_5_Factors',0],
        ['Europe','FF5','yearly','Europe_5_Factors',1],
        ['Europe','FF5','daily','Europe_5_Factors_Daily',0],  
        ['Japan','FF5','monthly','Japan_5_Factors',0],
        ['Japan','FF5','yearly','Japan_5_Factors',1],
        ['Japan','FF5','daily','Japan_5_Factors_Daily',0],    
        ['Asia_Pacific_ex_Japan','FF5','monthly','Asia_Pacific_ex_Japan_5_Factors',0],
        ['Asia_Pacific_ex_Japan','FF5','yearly','Asia_Pacific_ex_Japan_5_Factors',1],
        ['Asia_Pacific_ex_Japan','FF5','daily','Asia_Pacific_ex_Japan_5_Factors_Daily',0],   
        ['North_America','FF5','monthly','North_America_5_Factors',0],
        ['North_America','FF5','yearly','North_America_5_Factors',1],
        ['North_America','FF5','daily','North_America_5_Factors_Daily',0], \

        ['Global','Mom','monthly','Global_Mom_Factor',0],
        ['Global','Mom','yearly','Global_Mom_Factor',1],
        ['Global','Mom','daily','Global_Mom_Factor_Daily',0],   
        ['Global_ex_US','Mom','monthly','Global_ex_US_Mom_Factor',0],
        ['Global_ex_US','Mom','yearly','Global_ex_US_Mom_Factor',1],
        ['Global_ex_US','Mom','daily','Global_ex_US_Mom_Factor_Daily',0],  
        ['Europe','Mom','monthly','Europe_Mom_Factor',0],
        ['Europe','Mom','yearly','Europe_Mom_Factor',1],
        ['Europe','Mom','daily','Europe_Mom_Factor_Daily',0],  
        ['Japan','Mom','monthly','Japan_Mom_Factor',0],
        ['Japan','Mom','yearly','Japan_Mom_Factor',1],
        ['Japan','Mom','daily','Japan_Mom_Factor_Daily',0],    
        ['Asia_Pacific_ex_Japan','Mom','monthly','Asia_Pacific_ex_Japan_MOM_Factor',0],
        ['Asia_Pacific_ex_Japan','Mom','yearly','Asia_Pacific_ex_Japan_MOM_Factor',1],
        ['Asia_Pacific_ex_Japan','Mom','daily','Asia_Pacific_ex_Japan_MOM_Factor_Daily',0],   
        ['North_America','Mom','monthly','North_America_Mom_Factor',0],
        ['North_America','Mom','yearly','North_America_Mom_Factor',1],
        ['North_America','Mom','daily','North_America_Mom_Factor_Daily',0]                 
        ], columns=['scope','factor','freq','symbol','seq'])

   #数据源
   source='famafrench'
   if scope == "China":
       scope="Asia_Pacific_ex_Japan"
    
   #匹配：scope+factor+freq
   ss=s[s['scope'].isin([scope]) & s['factor'].isin([factor]) \
                                  & s['freq'].isin([freq])]  
   #如果未找到匹配的模式，显示信息后返回
   if len(ss)==0:
        print("  Sorry, no data available for",scope,factor,freq)
        return None

   #重新索引，第1行的索引编号为0
   sss=ss.reset_index(drop=True)    
   #取出对应的symbol
   symbol=sss.iloc[0]['symbol']    
   #取出对应的月(0)/年(1)编号
   seq=sss.iloc[0]['seq']

   #抓取数据
   import pandas_datareader.data as web
   try:
        ds = web.DataReader(symbol, source, start, end)
   except:
        print("  Sorry, connection to data source failed for",symbol)        
        return None
    
   #提取希望的资产定价因子
   factor_df=ds[seq]
   if len(factor_df)==0:
        print("  Sorry, no data available for",start,end,scope,factor,freq)
        #print("Tracing process:"s,ss,sss,symbol,seq,ds)
        return None    
    
   return factor_df

if __name__=='__main__':
    ff3_factors_us=get_ff_factors("01/01/2009","07/18/2019",scope='US', \
                                  factor='FF3',freq='yearly')

    ff3_factors_jp=get_ff_factors("01/01/2009","07/18/2019",scope='Japan', \
                                  factor='FF3',freq='yearly')

    ff3_factors_eu=get_ff_factors("2009-01-01","2019-07-18",scope='Europe', \
                                  factor='FF3',freq='yearly')



#==============================================================================
def get_ff3_factors(start,end,scope='US',freq='yearly'):    
    """
    功能：抓取Fama-French三因子模型的三个市场因子
    1）市场风险溢价(Rmarket-Rf，收益率风险溢价) 
    2）小市值风险溢价(SMB，规模因子，即小市值股票对比大市值股票的超额收益率)
    3）高市净率风险溢价(HML，价值因子，高溢价股票对比低溢价股票的超额收益率)
    
    输入参数：
    start/end：开始/结束日期，格式为YYYY-MM-DD或MM/DD/YYYY
    scope：地区(美国，日本，欧洲，全球，除美国外的全球，除日本外的亚太)
    freq：采样周期(日、周、月、年，其中周仅支持美国市场)
    输出：按照采样周期排列的指定地区的FF3因子(pandas格式)
    """
    
    #仅为测试用
    #start="2018-01-01"
    #end="2018-12-31"
    #scope='US'
    #freq='monthly'
    
    #抓取FF3因子
    factor='FF3'
    ff3_factors=get_ff_factors(start,end,scope,factor,freq)   
    
    return ff3_factors

if __name__=='__main__':
    factors=get_ff3_factors("2018-01-01","2018-12-31",scope='US', \
                                  freq='monthly')
    print("\n",factors)

#==============================================================================
def test_ff3_factors():
    """
    功能：交互式测试资产定价模型FF3的因子
    输入：用户输入
    显示：各个因子
    返回值：无
    """
    
    import easygui as g

    title="Test the Factors of Fama-French 3-factor Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Frequency (yearly, monthly, daily)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","US","monthly")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    """
    如果用户输入的值比选项少的话，则返回列表中的值用空字符串填充用户为输入的选项。
    如果用户输入的值比选项多的话，则返回的列表中的值将截断为选项的数量。
    如果用户取消操作，则返回域中的列表的值或者None值
    """
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        freq=fldvalues[3]
        
        ff_factors=get_ff3_factors(start,end,scope,freq)
        if ff_factors is None:
            g.msgbox(msg="Sorry, no factor data available under the condition.", \
                     title=title)                       
        else: 
            g.msgbox(msg=ff_factors,title=title)
    else:
        g.msgbox(msg="Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ff3_factors()        

#==============================================================================
def draw1_ff_factors(model,scope,factors,factor_type):    
    """
    功能：绘制单曲线的FF因子变化图
    输入参数：
    model: 模型类型, 任意字符串(例如FF3, FFC4, FF5)
    scope: 市场范围, 任意字符串(例如US, China, Europe)
    factors：包括FF各个因子的数据框
    factor_type：因子类型，严格限于RF/Mkt-RF/SMB/HML/MOM/RMW/CMA
    输出：图形
    返回值：无
    """
    #仅用作测试，完成后应注释掉
    #model="Fama-French 3-factor Model"
    #scope="US"
    #factor_type='ALL'
    
    if factor_type in ['Mkt-RF']:    
        #绘制市场收益率风险溢价变化图
        """
        线段的颜色：c='r'，红色
        线型：ls='-'，实线；':'则为虚线
        线段的粗细：lw=3
        节点形状：marker='o'，圆形
        节点大小：ms=10
        节点颜色：mfc='y'，黄色
        """
        try:
            factors.plot(y=['Mkt-RF'], \
                     c='r',ls='-',lw=3,marker='o',ms=10,mfc='y')
        except:
            print("  Sorry, no MKT factor data available")
            return
        plt.axhline(y=0.0,color='b',linestyle=':')  #画0点虚线X轴做基准
        title1="\n"+model+": "+scope+", "+" Mkt-RF Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('Mkt-RF',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8,rotation=45)
        plt.legend(loc='best')    
        plt.show()   
        
    if factor_type in ['SMB']:            
        #绘制规模因子变化图
        try:
            factors.plot(y=['SMB'], \
                     c='b',ls='-',lw=3,marker='o',ms=10,mfc='r')
        except:
            print("  Sorry, no SMB factor data available")
            return
        plt.axhline(y=0.0,color='b',linestyle=':')
        title1="\n"+model+": "+scope+", "+"SMB Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('SMB',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8,rotation=45)
        plt.legend(loc='best')
        plt.show()   

    if factor_type in ['HML']:     
        #绘制价值因子变化图
        try:
            factors.plot(y=['HML'], \
                     c='c',ls='-',lw=3,marker='o',ms=10,mfc='r')
        except:
            print("  Sorry, no HML factor data available")
            return
        plt.axhline(y=0.0,color='b',linestyle=':')  #画一条0点的虚线做基准参考
        title1="\n"+model+": "+scope+", "+"HML Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('HML',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8,rotation=45)
        plt.legend(loc='best')
        plt.show()   

    if factor_type in ['RF']:      
        #绘制无风险收益率变化图
        try:
            factors.plot(y=['RF'], \
                     c='g',ls='-',lw=3,marker='*',ms=14,mfc='r')
        except:
            print("  Sorry, no RF data available")
            return
        title1="\n"+model+": "+scope+", "+"HML Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('RF %',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8,rotation=45)
        plt.legend(loc='best')
        plt.show()  

    if factor_type in ['MOM']:         
        #绘制动量因子变化图
        try:
            factors.plot(y=['MOM'], \
                     c='deeppink',ls='-',lw=3,marker='*',ms=15,mfc='y')
        except:
            print("  Sorry, no MOM factor data available")
            return
        plt.axhline(y=0.0,color='b',linestyle=':')
        title1="\n"+model+": "+scope+", "+"MOM Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('Mom',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8)
        plt.legend(loc='best')
        plt.show()  


    if factor_type in ['RMW']:      
        #绘制盈利因子变化图
        try:
            factors.plot(y=['RMW'], \
                     c='blue',ls='-',lw=3,marker='*',ms=15,mfc='r')
        except:
            print("  Sorry, no RMW factor data available")
            return        
        plt.axhline(y=0.0,color='b',linestyle=':')
        title1="\n"+model+": "+scope+", "+"RMW Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('RMW',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8)
        plt.legend(loc='best')
        plt.show()  

    if factor_type in ['CMA']:          
        #绘制投资因子变化图
        try:
            factors.plot(y=['CMA'], \
                     c='black',ls='-',lw=3,marker='*',ms=15,mfc='r')
        except:
            print("  Sorry, no CMA factor data available")
            return
        plt.axhline(y=0.0,color='b',linestyle=':')
        title1="\n"+model+": "+scope+", "+"CMA Factor"
        plt.title(title1,fontsize=12,fontweight='bold')
        plt.ylabel('CMA',fontsize=12,fontweight='bold')
        plt.xticks(factors.index,fontsize=8)
        plt.legend(loc='best')
        plt.show() 
    
    return


if __name__=='__main__':
    factors1=get_ff3_factors("2018-01-01","2018-12-31",scope='US', \
                                  freq='monthly')
    factors2=get_ff3_factors("2018-01-01","2018-12-31",scope='China', \
                                  freq='monthly')    
    draw1_ff_factors("FF3 model","US",factors1,"SMB")
    draw1_ff_factors("FF3 model","Japan",factors2,"SMB")

#==============================================================================
def draw2_ff_factors(model,scope1,scope2,factors1,factors2,factor_type):    
    """
    功能：绘制双曲线的FF因子变化图
    输入参数：
    model: 模型类型, 任意字符串(例如FF3, FFC4, FF5)
    scope: 市场范围, 任意字符串(例如US, China, Europe)
    factors1/2：用于对比的包括FF各个因子的两个数据框
    factor_type：因子类型，严格限于RF/MKT/SMB/HML/MOM/RMW/CMA
    输出：图形
    返回值：无
    """
    #仅用作测试，完成后应注释掉
    #model="FF3 Model"
    #scope1="US"
    #scope2="Europe"
    #factor_type='Mkt-RF'
   
    #转换索引类型为DatetimeIndex，便于后续处理
    import pandas as pd
    factors1['Date']=factors1.index.strftime("%Y-%m-%d")
    factors1['Date']=pd.to_datetime(factors1['Date'])
    factors1.set_index('Date',inplace=True)
    
    factors2['Date']=factors2.index.strftime("%Y-%m-%d")
    factors2['Date']=pd.to_datetime(factors2['Date'])
    factors2.set_index('Date',inplace=True)    

    try:
        plt.plot(factors1[factor_type],label=scope1,marker='o')
        plt.plot(factors2[factor_type],label=scope2,marker='*')
    except:
        print("  Sorry, no available data for factor",factor_type)
        return
    plt.axhline(y=0.0,color='b',linestyle=':')  
    title1="\n"+model+": "+scope1+" vs. "+scope2+", "+" Factor "+factor_type
    plt.title(title1,fontsize=12,fontweight='bold')
    plt.ylabel(factor_type,fontsize=12,fontweight='bold')
    plt.legend(loc='best')    
    plt.show()       
    
    return

if __name__=='__main__':
    factors1=get_ff3_factors("2009-01-01","2018-12-31",scope='US', \
                                  freq='yearly')
    factors2=get_ff3_factors("2009-01-01","2018-12-31",scope='Europe', \
                                  freq='yearly')
    draw2_ff_factors("FF3 model","US","Europe",factors1,factors2,"Mkt-RF")
    draw2_ff_factors("FF3 model","US","Europe",factors1,factors2,"SMB")
    draw2_ff_factors("FF3 model","US","Europe",factors1,factors2,"HML")

#==============================================================================
def get_ffc4_factors(start,end,scope='US',freq='yearly'):    
    """
    功能：抓取Fama-French-Carhart四因子模型的四个市场因子
    1）市场风险溢价(Rmarket-Rf，收益率风险溢价) 
    2）小市值风险溢价(SMB，规模因子，即小市值股票对比大市值股票的超额收益率)
    3）高市净率风险溢价(HML，价值因子，高溢价股票对比低溢价股票的超额收益率)
    4）交易动量因子(MOM，市场交易中资本利差最高的那部分股票被称为赢家winner，
    而最低的那部分股票被称为输家loser。短期内，赢家具有保持作为赢家的趋势，
    输家具有继续作为输家的趋势。这种现象称为交易惯性，惯性即动量。动量因子
    就是高收益率股票对比低收益率股票的超额收益率)
    
    输入参数：
    start/end：开始/结束日期，格式为YYYY-MM-DD或MM/DD/YYYY
    scope：地区(美国，日本，欧洲，全球，除美国外的全球，除日本外的亚太)
    freq：采样周期(日、周、月、年，其中周仅支持美国市场)
    输出：按照采样周期排列的指定地区的FFC4因子(pandas格式)
    """
    
    #仅为测试用
    #start="2018-01-01"
    #end="2018-12-31"
    #scope='US'
    #freq='monthly'
    
    #抓取FF3因子
    factor='FF3'
    ff3=get_ff_factors(start,end,scope,factor,freq)   
    
    #抓取Mom因子
    factor='Mom'
    try:
        Mom=get_ff_factors(start,end,scope,factor,freq)  
    except:
        print("  Sorry, connection to data source failed:-( Try later!")
        return
    
    #将动量数据列名一致化
    if scope == 'US':
        Mom=Mom.rename(columns={'Mom   ':'MOM'})    #针对US数据
    else:
        Mom=Mom.rename(columns={'WML':'MOM'})    #针对非US数据        

    #合成FF3+Mom因子
    import pandas as pd
    ffc4_factors=pd.merge(ff3,Mom,how='inner',left_index=True,right_index=True)
    
    return ffc4_factors


if __name__=='__main__':
    start="2018-01-01"
    end="2018-12-31"
    scope='Europe'
    freq="monthly"
    factors1=get_ffc4_factors("2009-01-01","2018-12-31",scope='US', \
                                  freq='yearly')
    factors2=get_ffc4_factors("2009-01-01","2018-12-31",scope='Japan', \
                                  freq='yearly')
    draw1_ff_factors("FFC4 model","US",factors1,"MOM")


#==============================================================================
def test_ffc4_factors():
    """
    功能：交互式测试资产定价模型FFC4的因子
    输入：用户输入
    显示：各个因子
    返回值：无
    """
    
    import easygui as g

    title="Test the Factors of Fama-French-Carhart 4-factor Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Frequency (yearly, monthly, daily)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","US","monthly")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    """
    如果用户输入的值比选项少的话，则返回列表中的值用空字符串填充用户为输入的选项。
    如果用户输入的值比选项多的话，则返回的列表中的值将截断为选项的数量。
    如果用户取消操作，则返回域中的列表的值或者None值
    """
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        freq=fldvalues[3]
        
        ff_factors=get_ffc4_factors(start,end,scope,freq)
        if ff_factors is None:
            g.msgbox(msg="  Sorry, no factor data available under the condition.", \
                     title=title)                       
        else: 
            g.msgbox(msg=ff_factors,title=title)
    else:
        g.msgbox(msg="  Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ffc4_factors()        


#==============================================================================
def get_ff5_factors(start,end,scope='US',freq='yearly'):    
    """
    功能：抓取Fama-French五因子模型的五个市场因子
    1）市场风险溢价(Rmarket-Rf，收益率风险溢价) 
    2）小市值风险溢价(SMB，规模因子，即小市值股票对比大市值股票的超额收益率)
    3）高市净率风险溢价(HML，价值因子，高溢价股票对比低溢价股票的超额收益率)
    4）盈利因子(RMW，盈利好和盈利差的多元化股票组合收益之间的差异。其中，盈利
    定义为年营业收入减去营业成本、利息费用、销售费用和管理费用后再除以t-1财年
    末的账面权益)
    5）投资因子(CMA：投资低和投资高的多元化股票组合收益之间的差异(投资高对长远
    未来有利，但将影响当年业绩。其中，投资定义为t-1财年的新增总资产除以t-2财年
    末的总资产)
    
    输入参数：
    start/end：开始/结束日期，格式为YYYY-MM-DD或MM/DD/YYYY
    scope：地区(美国，日本，欧洲，全球，除美国外的全球，除日本外的亚太)
    freq：采样周期(日、周、月、年，其中周仅支持美国市场)
    输出：按照采样周期排列的指定地区的FF5因子(pandas格式)
    """
    
    #仅为测试用
    #start="2018-01-01"
    #end="2018-12-31"
    #scope='US'
    #freq='monthly'
    
    #抓取FF5因子
    factor='FF5'
    ff5_factors=get_ff_factors(start,end,scope,factor,freq)   
    
    return ff5_factors

if __name__=='__main__':
    factors1=get_ff5_factors("2018-01-01","2018-12-31",scope='US', \
                                  freq='monthly')
    draw1_ff_factors("FFC4 model","US",factors1,"RMW")
    
    factors2=get_ff5_factors("2018-01-01","2018-12-31",scope='Europe', \
                                  freq='monthly')    
    draw1_ff_factors("FFC4 model","US",factors2,"CMA")
    draw2_ff_factors("FFC4 model","US","Europe",factors1,factors2,"RMW")    

#==============================================================================
def test_ff5_factors():
    """
    功能：交互式测试资产定价模型FF5的因子
    输入：用户输入
    显示：各个因子
    返回值：无
    """
    
    import easygui as g

    title="Test the Factors of Fama-French 5-factor Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Frequency (yearly, monthly, daily)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","US","monthly")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    """
    如果用户输入的值比选项少的话，则返回列表中的值用空字符串填充用户为输入的选项。
    如果用户输入的值比选项多的话，则返回的列表中的值将截断为选项的数量。
    如果用户取消操作，则返回域中的列表的值或者None值
    """
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        freq=fldvalues[3]
        
        ff_factors=get_ff5_factors(start,end,scope,freq)
        if ff_factors is None:
            g.msgbox(msg="  Sorry, no factor data available under the condition.", \
                     title=title)                       
        else: 
            g.msgbox(msg=ff_factors,title=title)
    else:
        g.msgbox(msg="  Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ff5_factors()   


#==============================================================================

def reg_ff3_betas(ticker,start,end,scope='US',graph=True):
    """
    功能：测试一只股票对于FF3各个因子的系数大小、符号方向和显著性
    输入：
    ticker: 股票代码
    start/end: 开始/结束日期，格式：YYYY-MM-DD或MM/DD/YYYY
    scope: 股票市场地域
    输出：FF3模型对于一只股票日收益率的回归结果
    """
    
    #仅为测试使用，过后应注释掉
    #scope='Japan'
    #ticker='9983.T'
    #start='01/01/2018'
    #end='06/30/2019'        

    #抓取每日股价
    price=get_price(ticker,start,end)
    
    #计算股票日收益率
    import pandas as pd
    ret=pd.DataFrame(price['Close'].pct_change()*100)
    ret=ret.dropna()
    #命名日收益率字段为dRet
    ret=ret.rename(columns={'Close':'dRet'})

    #抓取FF3因子
    freq='daily'
    ff3=get_ff3_factors(start,end,scope,freq)
    if ff3 is None:
        print("  #Error(reg_ff3_betas): factors tentatively inaccessible for",scope,freq)
        return None
    
    #改造索引类型为时间戳索引（DatetimeIndex），便于与ret合成
    ff3['Date']=ff3.index.strftime("%m/%d/%Y")
    ff3['Date']=pd.to_datetime(ff3['Date'])
    ff3.set_index('Date',inplace=True)

    #内连接：股票日收益率+每日FF3因子
    sample=pd.merge(ret,ff3,how='inner',left_index=True,right_index=True)
    sample=sample.dropna()

    #回归FF3模型
    import statsmodels.api as sm
    y=sample['dRet']
    X=sample[['Mkt-RF','SMB','HML']]
    X1=sm.add_constant(X)
    results=sm.OLS(y,X1).fit() 
    
    #生成回归参数
    #dir(results)    #查看回归结果results的属性
    parms=regparms(results)

    #输出结果并绘制横向柱状图    
    if graph == True:
        gparms=parms.iloc[[1,2,3]]
        print("\n",parms)
        title=codetranslate(ticker)+": FF3模型的贝塔系数"
        plt.title(title,fontsize=12,fontweight='bold')
        plt.ylabel("贝塔系数",fontsize=12,fontweight='bold')
        
        import datetime; today = datetime.date.today()
        footnote="FF3因子"+"\n数据来源：雅虎财经, "+str(today)
        plt.xlabel(footnote,fontsize=12,fontweight='bold')
        plt.bar(gparms.index,gparms.coef,0.5,color=['r','g','b'],alpha=0.5)
        plt.axhline(y=0.0,color='blue',linestyle=':')

    return parms

if __name__=='__main__':
    reg_ff3_betas=reg_ff3_betas('9983.T',"2018-01-01","2018-12-31",scope='Japan')
    
#==============================================================================
def test_ff3_betas():
    """
    功能：交互式测试资产定价因子的贝塔系数
    输入：用户输入
    显示：各个因子的贝塔系数
    返回值：无
    """
    
    import easygui as g

    title="Calculate A Stock's Betas Using FF3 Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Stock code (mind the suffix for no-US stocks)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","US","AAPL")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        if scope == "China": scope="Asia_Pacific_ex_Japan"
        ticker=fldvalues[3]
        
        parms=reg_ff3_betas(ticker,start,end,scope)
        if parms is None:
            g.msgbox(msg="  Sorry, no beta data available under the condition.", \
                     title=title)                       
        else: 
            title=ticker+"'s Betas Using FF3 Model"
            g.msgbox(msg=parms,title=title)
    else:
        g.msgbox(msg="  Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ff3_betas()   

    
#==============================================================================
def reg_ffc4_betas(ticker,start,end,scope='US',graph=True):
    """
    功能：测试一只股票对于FFC4各个因子的系数大小、符号方向和显著性
    输入：
    ticker: 股票代码
    start/end: 开始/结束日期，格式：YYYY-MM-DD或MM/DD/YYYY
    scope: 股票市场地域
    输出：FFC4模型对于一只股票日收益率的回归结果
    """
    
    #仅为测试使用，过后应注释掉
    #scope='US'
    #ticker='AAPL'
    #start='01/01/2018'
    #end='06/30/2019'        

    #抓取每日股价
    import siat.security_prices as ssp
    price=get_price(ticker,start,end)
    
    #计算股票日收益率
    import pandas as pd
    ret=pd.DataFrame(price['Close'].pct_change()*100)
    ret=ret.dropna()
    #命名日收益率字段为dRet
    ret=ret.rename(columns={'Close':'dRet'})

    #抓取每日FFC4因子
    freq='daily'
    ffc4=get_ffc4_factors(start,end,scope,freq)
    if ffc4 is None:
        print("  #Error(reg_ffc4_betas): factors tentatively inaccessible for",scope,freq)
        return None    
    
    #改造索引类型为时间戳索引（DatetimeIndex），便于与ret合成
    ffc4['Date']=ffc4.index.strftime("%m/%d/%Y")
    ffc4['Date']=pd.to_datetime(ffc4['Date'])
    ffc4.set_index('Date',inplace=True)    

    #合成股票日收益率+每日FF3因子
    sample=pd.merge(ret,ffc4,how='inner',left_index=True,right_index=True)
    sample=sample.dropna()

    #回归FF3模型
    import statsmodels.api as sm
    y=sample['dRet']
    X=sample[['Mkt-RF','SMB','HML','MOM']]
    X1=sm.add_constant(X)
    results=sm.OLS(y,X1).fit() 
    
    #生成回归参数
    parms=regparms(results)

    #输出结果并绘制横向柱状图
    gparms=parms.iloc[[1,2,3,4]]
    if graph == True:
        #print("\n",parms)
        title=codetranslate(ticker)+": FFC4模型的贝塔系数"
        plt.title(title,fontsize=12,fontweight='bold')
        plt.ylabel("贝塔系数",fontsize=12,fontweight='bold')
        
        import datetime; today = datetime.date.today()
        footnote="FFC4因子"+"\n数据来源：雅虎财经, "+str(today)        
        plt.xlabel(footnote,fontsize=12,fontweight='bold')
        plt.bar(gparms.index,gparms.coef,0.5,color=['r','g','b'],alpha=0.5)
        plt.axhline(y=0.0,color='blue',linestyle=':')

    return parms


if __name__=='__main__':
    ffc4=reg_ffc4_betas('AAPL',"2018-01-01","2019-06-30",scope='US')
    
#==============================================================================
def test_ffc4_betas():
    """
    功能：交互式测试资产定价因子的贝塔系数
    输入：用户输入
    显示：各个因子的贝塔系数
    返回值：无
    """
    
    import easygui as g

    title="Calculate A Stock's Betas Using FFC4 Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Stock code (mind the suffix for no-US stocks)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","Europe","BMW.DE")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        if scope == "China": scope="Asia_Pacific_ex_Japan"
        ticker=fldvalues[3]
        
        parms=reg_ffc4_betas(ticker,start,end,scope)
        if parms is None:
            g.msgbox(msg="  Sorry, no beta data available under the condition.", \
                     title=title)                       
        else: 
            title=ticker+"'s Betas Using FFC4 Model"
            g.msgbox(msg=parms,title=title)
    else:
        g.msgbox(msg="  Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ffc4_betas() 
    
#==============================================================================
def reg_ff5_betas(ticker,start,end,scope='US',graph=True):
    """
    功能：测试一只股票对于FF5各个因子的系数大小、符号方向和显著性
    输入：
    ticker: 股票代码
    start/end: 开始/结束日期，格式：YYYY-MM-DD或MM/DD/YYYY
    scope: 股票市场地域
    输出：FF5模型对于一只股票日收益率的回归结果
    """
    
    #仅为测试使用，过后应注释掉
    #scope='US'
    #ticker='AAPL'
    #start='01/01/2018'
    #end='06/30/2019'        

    #抓取每日股价
    import siat.security_prices as ssp
    price=get_price(ticker,start,end)
    
    #计算股票日收益率
    import pandas as pd
    ret=pd.DataFrame(price['Close'].pct_change()*100)
    ret=ret.dropna()
    #命名日收益率字段为dRet
    ret=ret.rename(columns={'Close':'dRet'})

    #抓取每日FF3因子
    freq='daily'
    ff5=get_ff5_factors(start,end,scope,freq)
    if ff5 is None:
        print("  #Error(reg_ff5_betas): factors tentatively inaccessible for",scope,freq)
        return None       
    
    #改造索引类型为时间戳索引（DatetimeIndex），便于与ret合成
    ff5['Date']=ff5.index.strftime("%m/%d/%Y")
    ff5['Date']=pd.to_datetime(ff5['Date'])
    ff5.set_index('Date',inplace=True)    

    #合成股票日收益率+每日FF3因子
    sample=pd.merge(ret,ff5,how='inner',left_index=True,right_index=True)
    sample=sample.dropna()

    #回归FF3模型
    import statsmodels.api as sm
    y=sample['dRet']
    X=sample[['Mkt-RF','SMB','HML','RMW','CMA']]
    X1=sm.add_constant(X)
    results=sm.OLS(y,X1).fit() 
    
    #生成回归参数
    parms=regparms(results)

    #输出结果并绘制横向柱状图
    gparms=parms.iloc[[1,2,3,4,5]]
    if graph == True:
        #print("\n",parms)
        title=codetranslate(ticker)+"：FF5模型的贝塔系数"
        plt.title(title,fontsize=12,fontweight='bold')
        plt.ylabel("贝塔系数",fontsize=12,fontweight='bold')
        
        import datetime; today = datetime.date.today()
        footnote="FF5因子"+"\n数据来源：雅虎财经, "+str(today)           
        plt.xlabel(footnote,fontsize=12,fontweight='bold')
        plt.bar(gparms.index,gparms.coef,0.5,color=['r','g','b'],alpha=0.5)
        plt.axhline(y=0.0,color='blue',linestyle=':')

    return parms


if __name__=='__main__':
    ff5_betas=reg_ff5_betas('AAPL',"2018-01-01","2019-07-18",scope='US')
    
#==============================================================================
def test_ff5_betas():
    """
    功能：交互式测试资产定价因子的贝塔系数
    输入：用户输入
    显示：各个因子的贝塔系数
    返回值：无
    """
    
    import easygui as g

    title="Calculate A Stock's Betas Using FF5 Model"
    msg="Please fill in the information below"
    fldnames=["Start date (YYYY-MM-DD)","End date (YYYY-MM-DD)", \
          "Scope (US, China, Japan, Europe)", \
          "Stock code (mind the suffix for no-US stocks)"]
    fldvalues=[]
    values=("2018-01-01","2019-04-30","Japan","9983.T")
    fldvalues=g.multenterbox(msg,title=title,fields=fldnames,values=values)
    
    if fldvalues:
        start=fldvalues[0]
        end=fldvalues[1]
        scope=fldvalues[2]
        if scope == "China": scope="Asia_Pacific_ex_Japan"
        ticker=fldvalues[3]
        
        parms=reg_ff5_betas(ticker,start,end,scope)
        if parms is None:
            g.msgbox(msg="Sorry, no beta data available under the condition.", \
                     title=title)                       
        else: 
            title=ticker+"'s Betas Using FF5 Model"
            g.msgbox(msg=parms,title=title)
    else:
        g.msgbox(msg="Sorry, user cancelled the operation, no data retrieved.", \
                 title=title)
        
    return

if __name__=='__main__':
    test_ff5_betas() 
    
#==============================================================================    


# -*- coding: utf-8 -*-

def version():
    print("Version: 1.0.84")
    print("Security Investment Analysis Tool (SIAT)")
    print("Find a bug? Write to wdehong2000@163.com")
    print("Copyrighs (C) WANG Dehong, 2021")
    print("Not for commercial use unless authorised")

    
def author():
    print("Author: Dr. WANG Dehong, 王德宏")
    print("Professor, International Business School")
    print("Beijing Foreign Studies University")
    print("Email: wdehong2000@163.com")
    print("Room 618 International Business School Building, West Campus")
    print("No.19 West 3rd Ring North Road, Haidian District")
    print("Beijing, P. R. China. Zip code: 100089")


# -*- coding: utf-8 -*-
"""
版权：王德宏，北京外国语大学国际商学院
功能：
1、基于股票或股票组合计算无风险收益率
2、绘制无风险收益率的变化趋势图：日，周，月
3、与实际的无风险收益率比较
版本：1.0，2021-2-6
"""

#==============================================================================
#关闭所有警告
import warnings; warnings.filterwarnings('ignore')
import siat.common as com
from siat.translate import *
#==============================================================================
#==============================================================================
#==============================================================================
def compare_rf(rf1,col1,rf2,col2,fromdate,todate,power=0,zeroline=True,twinx=False):
    """
    功能：比较两个无风险收益率的时间序列，并绘制趋势线
    """
    #检查日期期间的合理性
    result,start,end=com.check_period(fromdate,todate)
    if not result:
        print(" Error(compare_rf): invalide date period from",fromdate,"to",todate)
        return None
    
    #检查并筛选两个无风险收益率的时间序列
    if rf1 is None:
        print(" Error(compare_rf): 1st risk-free-rate series is empty")
        return None        
    if rf2 is None:
        print(" Error(compare_rf): 2nd risk-free-rate series is empty")
        return None  
    df1a=rf1[rf1.index >= start]
    df1b=df1a[df1a.index <= end]
    df1b[col1+'%']=df1b[col1]*100.0
    
    df2a=rf2[rf2.index >= start]
    df2b=df2a[df2a.index <= end]
    df2b[col2+'%']=df2b[col2]*100.0

    #绘制对比图
    import siat.grafix as g
    ticker1=g.codetranslate(df1b['ticker'][0])
    colname1=col1+'%'
    label1=col1+'%'
    ticker2=g.codetranslate(df2b['ticker'][0])
    colname2=col2+'%'
    label2=col2+'%'   
    ylabeltxt='无风险收益率%'
    titletxt="基于CAPM计算的无风险收益率变化趋势"
    
    import datetime; today = datetime.date.today()
    rf1note=df1b['footnote'][0]
    rf2note=df2b['footnote'][0]
    footnote="无风险收益率1："+str(rf1note)+ \
             "\n无风险收益率2："+str(rf2note)+ \
             "\n数据来源：雅虎财经, "+str(today)
    
    g.plot_line2(df1b,ticker1,colname1,label1, \
               df2b,ticker2,colname2,label2, \
               ylabeltxt,titletxt,footnote, \
               power=power,zeroline=zeroline,twinx=twinx)

    return

if __name__=='__main__':
    df1=get_rf_capm('AAPL','^GSPC','2018-1-1','2020-12-31',window=40)
    df2=get_rf_capm('MSFT','^GSPC','2018-1-1','2020-12-31',window=40)
    compare_rf(df1,'Rf',df2,'Rf','2019-1-1','2019-12-31')
    compare_rf(df1,'Rf',df2,'Rf','2019-1-1','2019-12-31',twinx=True)

    df1['Rf_20']=df1['Rf'].rolling(window=20).mean()
    compare_rf(df1,'Rf',df1,'Rf_20','2019-1-1','2019-12-31')

#==============================================================================

def calc_rolling_cumret(dfc,col,period='Weekly'):
    """
    传入日收益率col的数据表dfc
    传出不同期间的累计收益率序列cumret
    """
    df=dfc.copy()
    #检查period类型
    periodlist = ["Weekly","Biweekly","Monthly","Quarterly","Semiannual","Annual"]
    if not (period in periodlist):
        print("  Error(calc_rolling_cumret): only supports:",periodlist)
        return None

    #换算期间对应的实际交易天数
    perioddays=[5,10,20,60,120,240]
    rollingnum=perioddays[periodlist.index(period)]      
    
    import numpy as np
    df['logdret']=np.log(df[col]+1)
    df['cumret']=np.exp(df['logdret'].rolling(rollingnum).sum())-1.0
    
    return df['cumret']

#==============================================================================

if __name__ =="__main__":
    col='Rf'
    limits=[0.01,0.01]
    
def winsor(df,col,limits=[0.01,0.01]):
    """
    功能：对于数据表df1中的列col进行下1%（参数1）和上1%（参数2）的处理
    """
    import numpy as np
    from scipy.stats.mstats import winsorize

    a = np.array(df[col])
    aw=winsorize(a,limits=limits)
    
    return aw.data
#==============================================================================
if __name__=='__main__':
    ticker='399001.SZ'
    mktidx='000300.SS'
    fromdate='2020-1-1'
    todate='2020-12-31'
    window=240

def get_rf_capm(ticker,mktidx,fromdate,todate,window=240,sharelist=[]):
    """
    功能：计算无风险收益率的时间序列
    ticker：股票或股票组合
    mktidx：股票市场指数
    fromdate：开始时间
    todate：截止时间
    window：每次回归的样本个数
    sharelist：第一个参数为投资组合的持股比例，默认为等权重；为单个股票时无用
    """
    #检查日期期间的合理性
    result,start,end=com.check_period(fromdate,todate)
    if not result:
        print(" Error(get_rf_capm): invalide date period from",fromdate,"to",todate)
        return None

    #提前开始日期，留出回归窗口
    start1=com.date_adjust(start,adjust=-window*2)
    
    #获得股票或股票组合的历史收益率
    import siat.security_prices as sp
    #单个股票情形
    if isinstance(ticker,str):
        spdf=sp.get_prices(ticker,start1,end)
    #股票组合情形
    if isinstance(ticker,list):
        if sharelist == []:
            num=len(ticker)
            sharelist=[1]*num
        spdf=sp.get_prices_portfolio(ticker,sharelist,start1,end)
    if spdf is None:
        print("  Error(get_rf_capm): info not found or unavailable for",ticker)
        return None
    
    spdf['Stock_dailyRet']=spdf['Close'].pct_change()  
    
    #获取市场指数的历史收益率
    rmdf=sp.get_prices(mktidx,start1,end)
    if rmdf is None:
        print("  Error(get_rf_capm): info not found or unavailable for",mktidx)
        return None    
    rmdf['Market_dailyRet']=rmdf['Close'].pct_change()  
    
    #合并股票（组合）与市场指数的收益率为一个数据集
    import pandas as pd
    df=pd.merge(rmdf['Market_dailyRet'],spdf['Stock_dailyRet'],how='inner',left_index=True,right_index=True)
    df['Date']=df.index.strftime("%Y-%m-%d")
    
    datelist_ts=list(df.index)    
    datelist=list(df['Date'])
    fromdate=start.strftime("%Y-%m-%d")
    #start_pos=lookup_datelist(datelist,fromdate,direction='more')
    todate=end.strftime("%Y-%m-%d")
    #end_pos=lookup_datelist(datelist,todate,direction='less')
    
    #用于滚动的日期期间
    datelist_rolling=[]
    for d in datelist:
        if (d >= fromdate) and (d <= todate):
            datelist_rolling=datelist_rolling+[d]
    
    #滚动回归
    if sharelist == []:
        #footnote=[ticker,mktidx,fromdate,todate,window]
        footnote=[ticker,mktidx,window]
    else:
        #footnote=[ticker,mktidx,fromdate,todate,window,sharelist]
        footnote=[ticker,mktidx,window,sharelist]
    betas=pd.DataFrame(columns=('date','Beta','alpha','R-sqr','p-value','sig','Rf','ticker','footnote'))    
    from scipy import stats
    import numpy as np
    
    for d in datelist_rolling:
        pos2=lookup_datelist(datelist,d)
        pos1=pos2 - window
        sdate2=datelist_ts[pos2]
        sdate1=datelist_ts[pos1]

        sampledf=df[df.index >= sdate1].copy()
        sampledf=sampledf[sampledf.index < sdate2]
        
        (beta,alpha,r_value,p_value,std_err)= \
            stats.linregress(sampledf['Market_dailyRet'],sampledf['Stock_dailyRet'])
        sig=sig_stars(p_value)
        try:
            rf=alpha/(1-beta)
        except:
            rf=np.NaN    
        
        row=pd.Series({'date':sdate2,'Beta':beta,'alpha':alpha, \
            'R-sqr':r_value**2,'p-value':p_value,'sig':sig,'Rf':rf,'ticker':ticker,'footnote':footnote})
        betas=betas.append(row,ignore_index=True)       
    
    betas.set_index('date',inplace=True)
    
    return betas

if __name__=='__main__':   
    df1=get_rf_capm('399001.SZ','000001.SS','2020-1-1','2020-12-31',window=240)
    df1['Rf'].plot()

    df2=get_rf_capm('000001.SS','399001.SZ','2020-1-1','2020-12-31',window=240)
    df2['Rf'].plot()

    df3=get_rf_capm('000001.SS','000300.SS','2020-1-1','2020-12-31',window=240)
    df3['Rf'].plot()    

    df4=get_rf_capm('AAPL','^GSPC','2020-1-1','2020-12-31',window=240)
    df4['Rf'].plot()  

    df5=get_rf_capm('^DJI','^GSPC','2020-1-1','2020-12-31',window=240)
    df5['Rf'].plot()  
#==============================================================================
if __name__=='__main__':   
    datelist=['2019-12-30','2019-12-31','2020-01-02','2020-01-03','2020-01-06']
    adate='2020-01-01'
    adate='2020-01-04'
    direction='less'
    direction='more'
    
def lookup_datelist(datelist,adate,direction='less'):
    """
    功能：在日期列表datelist查找与日期adate最接近日期的位置
    direction='more'：若无匹配的日期，则往日期增加的方向查找最接近日期的位置
    direction='less'：若无匹配的日期，则往日期减少的方向查找最接近日期的位置
    """
    i=0
    found=False
    while not found:
        try:
            pos=datelist.index(adate)
        except:
            if direction == 'more':
                i=i+1
            else:
                i=i-1
            adate=com.date_adjust(adate,adjust=i)
        else:
            found=True    

    return pos

#==============================================================================
if __name__=='__main__':   
    p_value=0.07
    p_value=0.02
    p_value=0.0009
    criteria='accounting'
    criteria='financial'
    
def sig_stars(p_value,criteria='accounting'):
    """
    功能：基于p_value给出星号的个数
    p_value：显著性水平
    criteria='accounting'：默认的显著性基准，<0.1为一颗星；若为'financial'，<0.05为一颗星
    """
    sig=''
    if criteria == 'accounting':
        if p_value < 0.1: sig='*'*1
        if p_value < 0.05: sig='*'*2
        if p_value < 0.01: sig='*'*3
    else:
        if p_value < 0.05: sig='*'*1
        if p_value < 0.01: sig='*'*2
        if p_value < 0.001: sig='*'*3        

    return sig

if __name__=='__main__': 
    sig_stars(0.1)
    sig_stars(0.07)
    sig_stars(0.04)
    sig_stars(0.009)

#==============================================================================
def calc_capm_rf(rmdf,rdf):
    """
    功能：CAPM回归
    返回：截距项，贝塔系数，无风险收益率
    """
    #OLS回归
    from scipy import stats 
    (beta,alpha,r_value,p_value,std_err)=stats.linregress(rmdf,rdf)    
    
    rf=alpha/(1-beta)
    
    return [beta,alpha,r_value,p_value,std_err,rf]

#==============================================================================
#==============================================================================
#==============================================================================
#==============================================================================
if __name__=='__main__': 
    start='2018-1-1'
    end='2020-12-31'
    scope='US'
    freq='daily'

def get_rf_kfdl(start,end,scope='US',freq='daily'):
    """
    功能：从Kenneth R. French's Data Library获得无风险收益率
    start/end：日期期间
    scope：国家/地区，支持美国/北美/欧洲/日本/不含日本的亚太/不含美国的全球。全球
    freq：支持日/月/年收益率，其中美国还支持周数据
    返回：无风险收益率，市场收益率
    """
    import siat.fama_french as ff
    factor='FF3'
    df=ff.get_ff_factors(start,end,scope,factor,freq)
    
    df['Market_dailyRet']=df['Mkt-RF']+df['RF']
    df['Rf']=df['RF']
    df['ticker']=scope
    footnote=[scope,freq]
    df['footnote']=df['ticker'].apply(lambda x:footnote)
    
    df1=df[['Market_dailyRet','Rf','ticker','footnote']]
    
    return df1


#==============================================================================
